import re

import validators
import subprocess
from six.moves.urllib.parse import urlparse
from validators import ValidationFailure

from netmeasure.measurements.base.measurements import BaseMeasurement
from netmeasure.measurements.file_download.results import FileDownloadMeasurementResult
from netmeasure.measurements.latency.measurements import LatencyMeasurement
from netmeasure.measurements.base.results import Error
from netmeasure.units import NetworkUnit, StorageUnit

WGET_OUTPUT_REGEX = re.compile(
    r"\((?P<download_rate>[\d.]*)\s(?P<download_unit>.*)\).*\[(?P<download_size>\d*)[\]/]"
)

WGET_ERRORS = {
    "wget-err": "wget had an unknown error.",
    "wget-split": "wget attempted to split the result but it was in an unanticipated format.",
    "wget-regex": "wget attempted get the known regex format and failed.",
    "wget-download-unit": "wget could not process the download unit.",
    "wget-download-rate": "wget could not process the download rate.",
    "wget-download-size": "wget could not process the download size.",
    "wget-no-server": "No closest server could be resolved.",
    "wget-timeout": "Measurement request timed out.",
}

WGET_DOWNLOAD_RATE_UNIT_MAP = {
    "KB/s": NetworkUnit("Kibit/s"),
    "MB/s": NetworkUnit("Mibit/s"),
}


class FileDownloadMeasurement(BaseMeasurement):
    """A measurement designed to test download speed."""

    def __init__(self, id, urls, count=4, download_timeout=180):
        """Initialisation of a download speed measurement.

        :param id: A unique identifier for the measurement.
        :param urls: A list of URLs used to perform the download speed
        measurement.
        :param count: A positive integer describing the number of
        pings to perform. Defaults to 4.
        :param download_timeout: An integer describing the number of
        seconds for the test to last. 0 means no timeout.
        """
        super(FileDownloadMeasurement, self).__init__(id=id)
        if len(urls) < 1:
            raise ValueError("At least one url must be provided.")
        for url in urls:
            validated_url = validators.url(url)
            if isinstance(validated_url, ValidationFailure):
                raise ValueError("`{url}` is not a valid url".format(url=url))

        if count < 0:
            raise ValueError(
                "A value of {count} was provided for the number of pings. This must be a positive "
                "integer or `0` to turn off the ping.".format(count=count)
            )

        if download_timeout < 0:
            raise ValueError(
                "A value of {count} was provided for the timeout. This must be a positive "
                "integer or `0` to turn off the timeout.".format(count=count)
            )

        self.urls = urls
        self.count = count
        self.download_timeout = download_timeout

    def measure(self):
        """Perform the measurement."""
        initial_latency_results = self._find_least_latent_url(self.urls)
        least_latent_url = initial_latency_results[0][0]
        results = [self._get_wget_results(least_latent_url, self.download_timeout)]
        if self.count > 0:
            host = urlparse(least_latent_url).netloc
            latency_measurement = LatencyMeasurement(self.id, host, count=self.count)
            results.append(latency_measurement.measure()[0])

        results.extend([res for _, res in initial_latency_results])
        return results

    def _find_least_latent_url(self, urls):
        """
        Performs a latency test for each specified endpoint
        Returns a sorted list of LatencyResults, sorted by average latency and None
        """
        initial_latency_results = []
        for url in urls:
            host = urlparse(url).netloc
            latency_measurement = LatencyMeasurement(self.id, host, count=2)
            initial_latency_results.append((url, latency_measurement.measure()[0]))
        return sorted(
            initial_latency_results,
            key=lambda x: (x[1].average_latency is None, x[1].average_latency),
        )

    def _get_wget_results(self, url, download_timeout):
        """Perform the download measurement."""
        if url is None:
            return self._get_wget_error("wget-no-server", url, traceback=None)

        if download_timeout == 0:
            download_timeout = None
        try:
            wget_out = subprocess.run(
                ["wget", "--tries=2", "-O", "/dev/null", url],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                timeout=download_timeout,
                universal_newlines=True,
            )
        except subprocess.TimeoutExpired:
            return self._get_wget_error("wget-timeout", url, traceback=None)

        if wget_out.returncode != 0:
            return self._get_wget_error("wget-err", url, traceback=wget_out.stderr)
        try:
            wget_data = wget_out.stderr.split("\n")[-3]
        except IndexError:
            return self._get_wget_error("wget-split", url, traceback=wget_out.stderr)
        matches = WGET_OUTPUT_REGEX.search(wget_data)

        try:
            match_data = matches.groupdict()
        except AttributeError:
            return self._get_wget_error("wget-regex", url, traceback=wget_out.stderr)

        if len(match_data.keys()) != 3:
            return self._get_wget_error("wget-regex", url, traceback=wget_out.stderr)

        try:
            download_rate_unit = WGET_DOWNLOAD_RATE_UNIT_MAP[
                match_data.get("download_unit")
            ]
        except KeyError:
            return self._get_wget_error(
                "wget-download-unit", url, traceback=wget_out.stderr
            )

        try:
            # NOTE: wget returns download rate in [K|M]B/s. Convert to [K|M]ibit/s.
            download_rate = float(match_data.get("download_rate")) * 8
        except (TypeError, ValueError):
            return self._get_wget_error(
                "wget-download-rate", url, traceback=wget_out.stderr
            )

        try:
            download_size = float(match_data.get("download_size"))
        except (TypeError, ValueError):
            return self._get_wget_error(
                "wget-download-size", url, traceback=wget_out.stderr
            )
        return FileDownloadMeasurementResult(
            id=self.id,
            url=url,
            download_rate_unit=download_rate_unit,
            download_rate=download_rate,
            download_size=download_size,
            download_size_unit=StorageUnit.bit,
            errors=[],
        )

    def _get_wget_error(self, key, url, traceback):
        return FileDownloadMeasurementResult(
            id=self.id,
            url=url,
            download_rate_unit=None,
            download_rate=None,
            download_size=None,
            download_size_unit=None,
            errors=[
                Error(
                    key=key, description=WGET_ERRORS.get(key, ""), traceback=traceback
                )
            ],
        )



import sys
import json
from unittest import TestCase, mock

import yt_dlp

from netmeasure.measurements.youtube_download.measurements import (
    YoutubeDownloadMeasurement,
    YOUTUBE_ERRORS,
)
from netmeasure.measurements.youtube_download.results import (
    YoutubeDownloadMeasurementResult,
)
from netmeasure.measurements.latency.results import LatencyMeasurementResult
from netmeasure.measurements.base.results import Error
from netmeasure.units import RatioUnit, TimeUnit, StorageUnit, NetworkUnit

"""
Note that the constructors for the yt_dlp errors occasionally add their own text in addition to the passed message!
"""


class YoutubeDownloadResultTestCase(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.id = "1"
        self.test_url = "https://www.youtube.com/watch?v=1233zthJUf31MA"
        self.ytm = YoutubeDownloadMeasurement(self.id, self.test_url)
        self.mock_progress_dicts = [
            {
                "_eta_str": "Unknown ETA",
                "_percent_str": "  0.0%",
                "_speed_str": "Unknown speed",
                "_total_bytes_str": "123.4MiB",
                "downloaded_bytes": 1024,
                "elapsed": 0.13415765762329102,
                "eta": None,
                "filename": "/tmp/youtube-dl_7954/1596695025_Portland Protest.mp4",
                "speed": None,
                "status": "downloading",
                "tmpfilename": "/tmp/youtube-dl_31918/1234567890_Fake Video.mp4.part",
                "total_bytes": 163492702,
            },
            {
                "_eta_str": "00:00",
                "_percent_str": "100.0%",
                "_speed_str": "12.34MiB/s",
                "_total_bytes_str": "123.4MiB",
                "downloaded_bytes": 163492702,
                "elapsed": 13.792589902877808,
                "eta": 0,
                "filename": "/tmp/youtube-dl_31918/1596679305_Portland Protest.mp4",
                "speed": 12345678.012500000,
                "status": "downloading",
                "tmpfilename": "/tmp/youtube-dl_31918/1234567890_Fake Video.mp4.part",
                "total_bytes": 163492702,
            },
            {
                "_elapsed_str": "12:34",
                "_total_bytes_str": "123.4MiB",
                "downloaded_bytes": 123456789,
                "elapsed": 12.345678987654321,
                "filename": "/tmp/youtube-dl_31918/1234567890_Fake Video.mp4",
                "status": "finished",
                "total_bytes": 123456789,
            },
        ]
        self.mock_progress_dicts_missing_attribute = [
            {
                "_eta_str": "Unknown ETA",
                "_percent_str": "  0.0%",
                "_speed_str": "Unknown speed",
                "_total_bytes_str": "123.4MiB",
                "downloaded_bytes": 1024,
                "elapsed": 0.13415765762329102,
                "eta": None,
                "filename": "/tmp/youtube-dl_7954/1596695025_Portland Protest.mp4",
                "speed": None,
                "status": "downloading",
                "tmpfilename": "/tmp/youtube-dl_31918/1234567890_Fake Video.mp4.part",
                "total_bytes": 163492702,
            },
            {
                "_eta_str": "00:00",
                "_percent_str": "100.0%",
                "_speed_str": "12.34MiB/s",
                "_total_bytes_str": "123.4MiB",
                "downloaded_bytes": 163492702,
                "elapsed": 13.792589902877808,
                "eta": 0,
                "filename": "/tmp/youtube-dl_31918/1596679305_Portland Protest.mp4",
                "speed": 12345678.012500000,
                "status": "downloading",
                "tmpfilename": "/tmp/youtube-dl_31918/1234567890_Fake Video.mp4.part",
                "total_bytes": 163492702,
            },
            {
                "_elapsed_str": "12:34",
                "_total_bytes_str": "123.4MiB",
                "downloaded_bytes": 123456789,
                "elapsed": 12.345678987654321,
                "filename": "/tmp/youtube-dl_31918/1234567890_Fake Video.mp4",
                "status": "finished",
                "BADotal_bytes": 123456789,
            },
        ]
        self.mock_progress_dicts_only_final = [
            {
                "_elapsed_str": "12:34",
                "_total_bytes_str": "123.4MiB",
                "downloaded_bytes": 123456789,
                "elapsed": 12.345678987654321,
                "filename": "/tmp/youtube-dl_31918/1234567890_Fake Video.mp4",
                "status": "finished",
                "total_bytes": 123456789,
            }
        ]
        self.mock_valid_youtube_download_result = YoutubeDownloadMeasurementResult(
            id=self.id,
            url=self.test_url,
            download_rate=98765424.1,
            download_rate_unit=NetworkUnit("bit/s"),
            download_size=123456789,
            download_size_unit=StorageUnit("B"),
            elapsed_time=12.345678987654321,
            elapsed_time_unit=TimeUnit("s"),
            errors=[],
        )
        self.mock_extraction_fail_result = YoutubeDownloadMeasurementResult(
            id=self.id,
            url=self.test_url,
            download_rate=None,
            download_rate_unit=None,
            download_size=None,
            download_size_unit=None,
            elapsed_time=None,
            elapsed_time_unit=None,
            errors=[
                Error(
                    key="youtube-extractor",
                    description=YOUTUBE_ERRORS.get("youtube-extractor", ""),
                    traceback="Extraction failed!; please report this issue on  https://github.com/yt-dlp/yt-dlp/issues?q= , filling out the appropriate issue template. Confirm you are on the latest version using  yt-dlp -U",
                )
            ],
        )
        self.mock_download_fail_result = YoutubeDownloadMeasurementResult(
            id=self.id,
            url=self.test_url,
            download_rate=None,
            download_rate_unit=None,
            download_size=None,
            download_size_unit=None,
            elapsed_time=None,
            elapsed_time_unit=None,
            errors=[
                Error(
                    key="youtube-download",
                    description=YOUTUBE_ERRORS.get("youtube-download", ""),
                    traceback="Download failed!",
                )
            ],
        )
        self.mock_missing_attribute_result = YoutubeDownloadMeasurementResult(
            id=self.id,
            url=self.test_url,
            download_rate=None,
            download_rate_unit=None,
            download_size=None,
            download_size_unit=None,
            elapsed_time=None,
            elapsed_time_unit=None,
            errors=[
                Error(
                    key="youtube-attribute",
                    description=YOUTUBE_ERRORS.get("youtube-attribute", ""),
                    traceback=str(self.mock_progress_dicts_missing_attribute),
                )
            ],
        )
        self.mock_final_only_result = (
            YoutubeDownloadMeasurementResult(
                id=self.id,
                url=self.test_url,
                download_rate=None,
                download_rate_unit=None,
                download_size=None,
                download_size_unit=None,
                elapsed_time=None,
                elapsed_time_unit=None,
                errors=[
                    Error(
                        key="youtube-progress_length",
                        description=YOUTUBE_ERRORS.get("youtube-progress_length", ""),
                        traceback=str(self.mock_progress_dicts_only_final),
                    )
                ],
            ),
        )
        self.mock_final_only_result = YoutubeDownloadMeasurementResult(
            id=self.id,
            url=self.test_url,
            download_rate=None,
            download_rate_unit=None,
            download_size=None,
            download_size_unit=None,
            elapsed_time=None,
            elapsed_time_unit=None,
            errors=[
                Error(
                    key="youtube-progress_length",
                    description=YOUTUBE_ERRORS.get("youtube-progress_length", ""),
                    traceback=str(self.mock_progress_dicts_only_final),
                )
            ],
        )
        self.mock_file_remove_result = YoutubeDownloadMeasurementResult(
            id=self.id,
            url=self.test_url,
            download_rate=None,
            download_rate_unit=None,
            download_size=None,
            download_size_unit=None,
            elapsed_time=None,
            elapsed_time_unit=None,
            errors=[
                Error(
                    key="youtube-file",
                    description=YOUTUBE_ERRORS.get("youtube-file", ""),
                    traceback="[Errno 2] No such file or directory: 'example_file'",
                )
            ],
        )
        self.mock_directory_remove_result = YoutubeDownloadMeasurementResult(
            id=self.id,
            url=self.test_url,
            download_rate=None,
            download_rate_unit=None,
            download_size=None,
            download_size_unit=None,
            elapsed_time=None,
            elapsed_time_unit=None,
            errors=[
                Error(
                    key="youtube-no_directory",
                    description=YOUTUBE_ERRORS.get("youtube-no_directory", ""),
                    traceback="[Errno 2] No such file or directory: 'example_dir'",
                )
            ],
        )
        self.mock_directory_remove_nonempty_result = YoutubeDownloadMeasurementResult(
            id=self.id,
            url=self.test_url,
            download_rate=None,
            download_rate_unit=None,
            download_size=None,
            download_size_unit=None,
            elapsed_time=None,
            elapsed_time_unit=None,
            errors=[
                Error(
                    key="youtube-directory_nonempty",
                    description=YOUTUBE_ERRORS.get("youtube-directory_nonempty", ""),
                    traceback="[Errno 39] Directory not empty: 'example_dir'",
                )
            ],
        )

    @mock.patch("shutil.rmtree")
    @mock.patch("os.remove")
    @mock.patch("os.rmdir")
    @mock.patch.object(yt_dlp, "YoutubeDL")
    def test_youtube_download_result_valid(
        self, mock_YoutubeDL, mock_rmdir, mock_remove, mock_rmtree
    ):
        self.ytm.progress_dicts = self.mock_progress_dicts
        mock_ydl = mock.MagicMock()
        mock_ydl.extract_info.return_value = None
        mock_YoutubeDL.return_value = mock_ydl
        mock_rmdir.side_effect = [0]
        mock_remove.side_effect = [0]
        mock_rmtree.side_effect = [0]
        self.assertEqual(
            self.ytm._get_youtube_download_result(self.test_url),
            self.mock_valid_youtube_download_result,
        )

    @mock.patch("shutil.rmtree")
    @mock.patch("os.remove")
    @mock.patch("os.rmdir")
    @mock.patch.object(yt_dlp, "YoutubeDL")
    def test_youtube_download_result_extraction_error(
        self, mock_YoutubeDL, mock_rmdir, mock_remove, mock_rmtree
    ):
        self.ytm.progress_dicts = self.mock_progress_dicts
        mock_ydl = mock.MagicMock()
        mock_ydl.extract_info.side_effect = [
            yt_dlp.utils.ExtractorError("Extraction failed!")
        ]
        mock_YoutubeDL.return_value = mock_ydl
        mock_rmdir.side_effect = [0]
        mock_remove.side_effect = [0]
        mock_rmtree.side_effect = [0]

        # print(self.ytm._get_youtube_download_result(self.test_url))
        # print(self.mock_extraction_fail_result)

        self.assertEqual(
            self.ytm._get_youtube_download_result(self.test_url),
            self.mock_extraction_fail_result,
        )

    @mock.patch("shutil.rmtree")
    @mock.patch("os.remove")
    @mock.patch("os.rmdir")
    @mock.patch.object(yt_dlp, "YoutubeDL")
    def test_youtube_download_result_download_error(
        self, mock_YoutubeDL, mock_rmdir, mock_remove, mock_rmtree
    ):
        self.ytm.progress_dicts = self.mock_progress_dicts
        mock_ydl = mock.MagicMock()
        mock_ydl.extract_info.side_effect = [
            yt_dlp.utils.DownloadError("Download failed!")
        ]
        mock_YoutubeDL.return_value = mock_ydl
        mock_rmdir.side_effect = [0]
        mock_remove.side_effect = [0]
        mock_rmtree.side_effect = [0]
        self.assertEqual(
            self.ytm._get_youtube_download_result(self.test_url),
            self.mock_download_fail_result,
        )

    @mock.patch("shutil.rmtree")
    @mock.patch("os.remove")
    @mock.patch("os.rmdir")
    @mock.patch.object(yt_dlp, "YoutubeDL")
    def test_youtube_download_result_attribute_error(
        self, mock_YoutubeDL, mock_rmdir, mock_remove, mock_rmtree
    ):
        self.ytm.progress_dicts = self.mock_progress_dicts_missing_attribute
        mock_ydl = mock.MagicMock()
        mock_ydl.extract_info.side_effect = None
        mock_YoutubeDL.return_value = mock_ydl
        mock_rmdir.side_effect = [0]
        mock_remove.side_effect = [0]
        mock_rmtree.side_effect = [0]
        self.assertEqual(
            self.ytm._get_youtube_download_result(self.test_url),
            self.mock_missing_attribute_result,
        )

    @mock.patch("shutil.rmtree")
    @mock.patch("os.remove")
    @mock.patch("os.rmdir")
    @mock.patch.object(yt_dlp, "YoutubeDL")
    def test_youtube_download_result_only_final_progress(
        self, mock_YoutubeDL, mock_rmdir, mock_remove, mock_rmtree
    ):
        self.ytm.progress_dicts = self.mock_progress_dicts_only_final
        mock_ydl = mock.MagicMock()
        mock_ydl.extract_info.side_effect = None
        mock_YoutubeDL.return_value = mock_ydl
        mock_rmdir.side_effect = [0]
        mock_remove.side_effect = [0]
        mock_rmtree.side_effect = [0]
        self.assertEqual(
            self.ytm._get_youtube_download_result(self.test_url),
            self.mock_final_only_result,
        )

    # NOTE: OUTDATED
    # @mock.patch("shutil.rmtree")
    # @mock.patch("os.remove")
    # @mock.patch("os.rmdir")
    # @mock.patch.object(yt_dlp, "YoutubeDL")
    # def test_youtube_download_result_remove_file(self, mock_YoutubeDL, mock_rmdir, mock_remove, mock_rmtree):
    #     self.ytm.progress_dicts = self.mock_progress_dicts
    #     mock_ydl = mock.MagicMock()
    #     mock_ydl.extract_info.side_effect = None
    #     mock_YoutubeDL.return_value = mock_ydl
    #     mock_rmdir.side_effect = [0]
    #     mock_remove.side_effect = [0]
    #     mock_rmtree.side_effect = [
    #         FileNotFoundError("[Errno 2] No such file or directory: 'example_file'")
    #     ]
    #     self.assertEqual(
    #         self.ytm._get_youtube_download_result(self.test_url), self.mock_file_remove_result
    #     )

    @mock.patch("shutil.rmtree")
    @mock.patch("os.remove")
    @mock.patch("os.rmdir")
    @mock.patch.object(yt_dlp, "YoutubeDL")
    def test_youtube_download_result_remove_directory(
        self, mock_YoutubeDL, mock_rmdir, mock_remove, mock_rmtree
    ):
        self.ytm.progress_dicts = self.mock_progress_dicts
        mock_ydl = mock.MagicMock()
        mock_ydl.extract_info.side_effect = None
        mock_YoutubeDL.return_value = mock_ydl
        mock_rmtree.side_effect = [
            FileNotFoundError("[Errno 2] No such file or directory: 'example_dir'")
        ]
        mock_rmdir.side_effect = [0]
        mock_remove.side_effect = [0]
        self.assertEqual(
            self.ytm._get_youtube_download_result(self.test_url),
            self.mock_directory_remove_result,
        )

    # NOTE: OUTDATED
    # @mock.patch("os.remove")
    # @mock.patch("os.rmdir")
    # @mock.patch.object(yt_dlp, "YoutubeDL")
    # def test_youtube_download_result_remove_directory_nonempty(
    #     self, mock_YoutubeDL, mock_rmdir, mock_remove
    # ):
    #     self.ytm.progress_dicts = self.mock_progress_dicts
    #     mock_ydl = mock.MagicMock()
    #     mock_ydl.extract_info.side_effect = None
    #     mock_YoutubeDL.return_value = mock_ydl
    #     mock_rmdir.side_effect = [
    #         OSError("[Errno 39] Directory not empty: 'example_dir'")
    #     ]
    #     mock_remove.side_effect = [0]
    #     self.assertEqual(
    #         self.ytm._get_youtube_download_result(self.test_url),
    #         self.mock_directory_remove_nonempty_result,
    #     )



import typing
from dataclasses import dataclass

from netmeasure.measurements.base.results import MeasurementResult
from netmeasure.units import TimeUnit, StorageUnit, RatioUnit, NetworkUnit


@dataclass(frozen=True)
class SpeedtestDotnetMeasurementResult(MeasurementResult):
    """Encapsulates the results from a speedtest_dotnet measurement.

    :param download_rate: The measured download rate.
    :param download_rate_unit: The unit of measurement of `download_rate`.
    :param upload_rate: The measured upload rate.
    :param upload_rate_unit: The unit of measurement of `upload_rate`.
    :param data_received: The quantity of data report by the speedtest utility
    :param data_received_unit: The unit of measurement of `data_received`
    :param latency: The measured latency.
    :param server_name: The name of the speedtest.net server used to perform
    the speedtest_dotnet measurement.
    :param server_id: The id of the speedtest.net server used to perform the
    speedtest_dotnet measurement.
    :param server_sponsor: The sponsor of the speedtest.net server used to
    perform the speedtest_dotnet measurement.
    :param server_host: The host name of the speedtest.net server used to
    perform the speedtest_dotnet measurement.
    """

    download_rate: typing.Optional[float]
    download_rate_unit: typing.Optional[NetworkUnit]
    upload_rate: typing.Optional[float]
    upload_rate_unit: typing.Optional[NetworkUnit]
    data_received: typing.Optional[float]
    data_received_unit: typing.Optional[StorageUnit]
    latency: typing.Optional[float]
    server_name: typing.Optional[str]
    server_id: typing.Optional[str]
    server_sponsor: typing.Optional[str]
    server_host: typing.Optional[str]



import typing
from dataclasses import dataclass


@dataclass(frozen=True)
class Error:
    """An error format for use with `MeasurementResult`.

    This data class is designed to be used with the `MeasurementResult`
    data class and its' subclasses to describe any errors that
    occurred in a measurement.

    :param key: A key to describe the error type.
    :param description: A human readable description of the encountered
    error.
    :param traceback: The traceback or outputs of a command that caused
    the error to occur.
    """

    key: str
    description: str
    traceback: str


@dataclass(frozen=True)
class MeasurementResult:
    """ "A standard interface for measurement results

    :param id: A unique identifier for the measurement result.
    :param errors: The errors that occurred while attempting to take
    the measurement.
    """

    id: str
    errors: typing.List[Error]



"""
Honesty Box Measurement

A framework for measuring things and producing structured results.
Copyright (C) 2019 Honesty Box Engineering
engineering@honestybox.com.au

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""


import typing

from netmeasure.measurements.base.results import MeasurementResult


class BaseMeasurement:
    """Interface for creating measurements.

    This base class is designed to be sub-classed and used as an
    interface for creating new measurement types.
    """

    def __init__(self, id):
        """Initialisation of a base measurement

        :param id: A unique identifier for the measurement.
        """
        super(BaseMeasurement, self).__init__()
        self.id = id

    def measure(self):
        """Perform the measurement and return the measurement results."""
        raise NotImplementedError



# -*- coding: utf-8 -*-
from unittest import TestCase, mock
import six
import subprocess

from netmeasure.measurements.latency.measurements import LatencyMeasurement
from netmeasure.measurements.base.results import Error
from netmeasure.measurements.file_download.measurements import WGET_OUTPUT_REGEX
from netmeasure.measurements.file_download.measurements import FileDownloadMeasurement
from netmeasure.measurements.file_download.measurements import WGET_ERRORS
from netmeasure.measurements.file_download.results import FileDownloadMeasurementResult
from netmeasure.measurements.latency.results import LatencyMeasurementResult

from netmeasure.units import NetworkUnit, StorageUnit

# NOTE: To match what subprocess calls output, wget output strings
#       should end with "\n\n" and latency output strings should end with "\n"


def test_wget_output_regex_accepts_anticipated_format():
    anticipated_format = six.ensure_str(
        "2019-08-07 09:12:08 (16.7 MB/s) - '/dev/null’ saved [11376]"
    )
    results = WGET_OUTPUT_REGEX.search(anticipated_format).groupdict()
    assert results == {
        "download_rate": "16.7",
        "download_size": "11376",
        "download_unit": "MB/s",
    }


class FileDownloadMeasurementCreationTestCase(TestCase):
    def test_invalid_hosts(self, *args):
        self.assertRaises(
            ValueError, FileDownloadMeasurement, "test", ["invalid..host"]
        )

    def test_invalid_count(self, *args):
        self.assertRaises(
            TypeError,
            FileDownloadMeasurement,
            "test",
            ["http://validfakeurl.com"],
            count="invalid-count",
        )

    def test_invalid_negative_count(self, *args):
        self.assertRaises(
            ValueError,
            FileDownloadMeasurement,
            "test",
            ["http://validfakeurl.com"],
            count=-2,
        )


class FileDownloadMeasurementWgetTestCase(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.measurement = FileDownloadMeasurement(
            "test", ["https://validfakehost.com/test"]
        )
        self.valid_wget_kibit_sec = FileDownloadMeasurementResult(
            id="test",
            url="http://validfakehost.com/test",
            download_rate_unit=NetworkUnit("Kibit/s"),
            download_rate=133.6,
            download_size=11376,
            download_size_unit=StorageUnit.bit,
            errors=[],
        )
        self.valid_wget_mibit_sec = FileDownloadMeasurementResult(
            id="test",
            url="http://validfakehost.com/test",
            download_rate_unit=NetworkUnit("Mibit/s"),
            download_rate=133.6,
            download_size=11376,
            download_size_unit=StorageUnit.bit,
            errors=[],
        )
        self.invalid_wget_mibit_sec = FileDownloadMeasurementResult(
            id="test",
            url="http://validfakehost.com/test",
            download_rate_unit=None,
            download_rate=None,
            download_size=None,
            download_size_unit=None,
            errors=[
                Error(
                    key="wget-err",
                    description=WGET_ERRORS.get("wget-err", ""),
                    traceback="\n2019-08-07 09:12:08 (16.7 MB/s) - '/dev/null’ saved [11376]\n\n",
                )
            ],
        )
        self.invalid_wget_download_unit = FileDownloadMeasurementResult(
            id="test",
            url="http://validfakehost.com/test",
            download_rate_unit=None,
            download_rate=None,
            download_size=None,
            download_size_unit=None,
            errors=[
                Error(
                    key="wget-download-unit",
                    description=WGET_ERRORS.get("wget-download-unit", ""),
                    traceback="\n2019-08-07 09:12:08 (16.7 TB/s) - '/dev/null’ saved [11376]\n\n",
                )
            ],
        )
        self.invalid_regex = FileDownloadMeasurementResult(
            id="test",
            url="http://validfakehost.com/test",
            download_rate_unit=None,
            download_rate=None,
            download_size=None,
            download_size_unit=None,
            errors=[
                Error(
                    key="wget-regex",
                    description=WGET_ERRORS.get("wget-regex", ""),
                    traceback="\n2019-08-07 09:12:08 [BAD REGEX]\n\n",
                )
            ],
        )

    @mock.patch("subprocess.run")
    def test_valid_wget_kibit_sec(self, mock_run):
        mock_run.return_value = subprocess.CompletedProcess(
            args=[],
            returncode=0,
            stdout="b''",
            stderr="\n2019-08-07 09:12:08 (16.7 KB/s) - '/dev/null’ saved [11376]\n\n",
        )
        self.assertEqual(
            self.valid_wget_kibit_sec,
            self.measurement._get_wget_results(
                "http://validfakehost.com/test", self.measurement.download_timeout
            ),
        )

    @mock.patch("subprocess.run")
    def test_valid_wget_mibit_sec(self, mock_run):
        mock_run.return_value = subprocess.CompletedProcess(
            args=[],
            returncode=0,
            stdout="b''",
            stderr="\n2019-08-07 09:12:08 (16.7 MB/s) - '/dev/null’ saved [11376]\n\n",
        )
        self.assertEqual(
            self.valid_wget_mibit_sec,
            self.measurement._get_wget_results(
                "http://validfakehost.com/test", self.measurement.download_timeout
            ),
        )

    @mock.patch("subprocess.run")
    def test_invalid_wget(self, mock_run):
        mock_run.return_value = subprocess.CompletedProcess(
            args=[],
            returncode=1,
            stdout="b''",
            stderr="\n2019-08-07 09:12:08 (16.7 MB/s) - '/dev/null’ saved [11376]\n\n",
        )
        self.assertEqual(
            self.invalid_wget_mibit_sec,
            self.measurement._get_wget_results(
                "http://validfakehost.com/test", self.measurement.download_timeout
            ),
        )

    @mock.patch("subprocess.run")
    def test_invalid_wget_download_unit(self, mock_run):
        mock_run.return_value = subprocess.CompletedProcess(
            args=[],
            returncode=0,
            stdout="b''",
            stderr="\n2019-08-07 09:12:08 (16.7 TB/s) - '/dev/null’ saved [11376]\n\n",
        )
        self.assertEqual(
            self.invalid_wget_download_unit,
            self.measurement._get_wget_results(
                "http://validfakehost.com/test", self.measurement.download_timeout
            ),
        )

    @mock.patch("subprocess.run")
    def test_wget_invalid_regex(self, mock_run):
        mock_run.return_value = subprocess.CompletedProcess(
            args=[],
            returncode=0,
            stdout="b''",
            stderr="\n2019-08-07 09:12:08 [BAD REGEX]\n\n",
        )
        self.assertEqual(
            self.invalid_regex,
            self.measurement._get_wget_results(
                "http://validfakehost.com/test", self.measurement.download_timeout
            ),
        )


class FileDownloadMeasurementClosestServerTestCase(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.example_urls = [
            "http://n1-validfakehost.com",
            "http://n2-validfakehost.com",
            "http://n3-validfakehost.com",
        ]
        self.measurement = FileDownloadMeasurement("test", self.example_urls)
        print("asdf")

    @mock.patch.object(LatencyMeasurement, "measure")
    def test_sort_least_latent_url(self, mock_latency_results):
        results = [
            (
                LatencyMeasurementResult(
                    id="test",
                    host="n1-validfakehost.com",
                    minimum_latency=None,
                    average_latency=None,
                    maximum_latency=None,
                    median_deviation=None,
                    errors=[],
                    packets_transmitted=None,
                    packets_received=None,
                    packets_lost=None,
                    packets_lost_unit=None,
                    elapsed_time=None,
                    elapsed_time_unit=None,
                ),
            ),
            (
                LatencyMeasurementResult(
                    id="test",
                    host="n2-validfakehost.com",
                    minimum_latency=None,
                    average_latency=25.0,
                    maximum_latency=None,
                    median_deviation=None,
                    errors=[],
                    packets_transmitted=None,
                    packets_received=None,
                    packets_lost=None,
                    packets_lost_unit=None,
                    elapsed_time=None,
                    elapsed_time_unit=None,
                ),
            ),
            (
                LatencyMeasurementResult(
                    id="test",
                    host="n3-validfakehost.com",
                    minimum_latency=None,
                    average_latency=999.0,
                    maximum_latency=None,
                    median_deviation=None,
                    errors=[],
                    packets_transmitted=None,
                    packets_received=None,
                    packets_lost=None,
                    packets_lost_unit=None,
                    elapsed_time=None,
                    elapsed_time_unit=None,
                ),
            ),
        ]
        mock_latency_results.side_effect = results
        self.assertEqual(
            self.measurement._find_least_latent_url(self.example_urls),
            [
                (self.example_urls[1], results[1][0]),
                (self.example_urls[2], results[2][0]),
                (self.example_urls[0], results[0][0]),
            ],
        )

    @mock.patch.object(LatencyMeasurement, "measure")
    def test_sort_one_url(self, mock_latency_results):
        results = [
            (
                LatencyMeasurementResult(
                    id="test",
                    host="n2-validfakehost.com",
                    minimum_latency=None,
                    average_latency=25.0,
                    maximum_latency=None,
                    median_deviation=None,
                    errors=[],
                    packets_transmitted=None,
                    packets_received=None,
                    packets_lost=None,
                    packets_lost_unit=None,
                    elapsed_time=None,
                    elapsed_time_unit=None,
                ),
            )
        ]
        mock_latency_results.side_effect = results
        self.assertEqual(
            self.measurement._find_least_latent_url([self.example_urls[1]]),
            [(self.example_urls[1], results[0][0])],
        )



import uuid

import click
from rich.console import Console
from rich.theme import Theme
from exitstatus import ExitStatus
from halo import Halo

from .measurements.file_download.measurements import FileDownloadMeasurement
from .measurements.file_download.results import FileDownloadMeasurementResult
from .measurements.ip_route.measurements import IPRouteMeasurement
from .measurements.ip_route.results import IPRouteMeasurementResult
from .measurements.latency.measurements import LatencyMeasurement
from .measurements.latency.results import LatencyMeasurementResult
from .measurements.netflix_fast.measurements import NetflixFastMeasurement
from .measurements.netflix_fast.results import NetflixFastMeasurementResult
from .measurements.netflix_fast.results import NetflixFastThreadResult
from .measurements.speedtest_dotnet.measurements import SpeedtestDotnetMeasurement
from .measurements.speedtest_dotnet.results import SpeedtestDotnetMeasurementResult
from .measurements.webpage_download.measurements import WebpageDownloadMeasurement
from .measurements.webpage_download.results import WebpageDownloadMeasurementResult
from .measurements.youtube_download.measurements import YoutubeDownloadMeasurement
from .measurements.youtube_download.results import YoutubeDownloadMeasurementResult


OUTPUT_THEME = Theme(
    {
        "info": "dim cyan",
        "warning": "magenta",
        "error": "bold red",
        "endpoint": "dim cyan",
        "value": "bold deep_sky_blue1",
        "unit": "white",
        "header": "bold white",
    }
)


def get_uuid_str() -> str:
    """
    Get a unique identifier string to assign to a measurement.
    """
    return str(uuid.uuid4())


@click.group(chain=True)
@click.option("-v", "--verbose", default=False, is_flag=True, required=False)
def cli(verbose):
    """
    Internet connection quality measurements.
    """
    pass


@cli.command("file_download")
@click.option(
    "-u", "--url", required=True, multiple=True, help="URL of file to download"
)
def perform_file_download_measurement(url):
    """
    Perform a file download measurement.

    Determines the URL with the lowest latency and then downloads it using wget.
    """

    console = Console(theme=OUTPUT_THEME)
    try:
        measurement = FileDownloadMeasurement(
            id=get_uuid_str(),
            urls=url,
        )
    except ValueError as err:
        raise click.BadParameter(err)
    with Halo(text="Performing File Download measurement", spinner="dots"):
        results = measurement.measure()
    output = f""
    for result in [r for r in results if type(r) == FileDownloadMeasurementResult]:
        if len(result.errors) > 0:
            for error in result.errors:
                console.print(f"[error]Error:[/error] {error.description}")
            return ExitStatus.failure
        output += (
            f"[header]:floppy_disk:  File Download   :floppy_disk:[/header]\n"
            f"URL: [endpoint]{result.url}[/endpoint]\n"
            f"Download Rate: [value]{result.download_rate}[/value] [unit]{result.download_rate_unit.value}[/unit] | "
            f"Download Size: [value]{result.download_size}[/value] [unit]{result.download_size_unit.value}[/unit]"
        )
    console.rule()
    console.print(output)
    console.rule()
    return ExitStatus.success


@cli.command("ip_route")
@click.option(
    "-h", "--host", required=True, multiple=True, help="Host to measure ip route to"
)
def perform_ip_route_measurement(host):
    """
    Perform an ip route measurement.

    Determines the host with the lowest latency and then measures the route to it.
    """

    console = Console(theme=OUTPUT_THEME)
    try:
        measurement = IPRouteMeasurement(
            id=get_uuid_str(),
            hosts=host,
        )
    except ValueError as err:
        raise click.BadParameter(err)
    with Halo(text="Performing IP Route measurement", spinner="dots"):
        results = measurement.measure()
    output = f""
    for result in [r for r in results if type(r) == IPRouteMeasurementResult]:
        if len(result.errors) > 0:
            for error in result.errors:
                console.print(f"[error]Error:[/error] {error.description}")
            return ExitStatus.failure
        output += (
            f"[header]:world_map:     IP Route     :world_map:[/header]\n"
            f"Host: [endpoint]{result.host}[/endpoint] | "
            f"IP: [endpoint]{result.ip}[/endpoint]\n"
            f"Route: [value]{', '.join(result.route)}[/value]\n"
            f"Hops: [value]{result.hop_count}[/value]"
        )
    console.rule()
    console.print(output)
    console.rule()
    return ExitStatus.success


@cli.command("latency")
@click.option(
    "-h", "--host", required=True, multiple=False, help="Host to measure latency to"
)
@click.option(
    "-c",
    "--count",
    default=3,
    required=False,
    multiple=False,
    help="Count of pings to send",
)
def perform_latency_measurement(host, count):
    """
    Perform a latency measurement.
    """

    console = Console(theme=OUTPUT_THEME)
    try:
        measurement = LatencyMeasurement(
            id=get_uuid_str(),
            host=host,
            count=count,
        )
    except ValueError as err:
        raise click.BadParameter(err)
    with Halo(text="Performing Latency measurement", spinner="dots"):
        results = measurement.measure()
    output = f"[header]:table_tennis_paddle_and_ball:     Latency     :table_tennis_paddle_and_ball:[/header]\n"
    for result in [r for r in results if type(r) == LatencyMeasurementResult]:
        if len(result.errors) > 0:
            for error in result.errors:
                console.print(f"[]Error:[/error] {error.description}")
                return ExitStatus.failure
        output += (
            f"Host: [endpoint]{result.host}[/endpoint]\n"
            f"Minimum Latency: [value]{result.minimum_latency}[/value] [unit]{result.elapsed_time_unit.value}[/unit] | "
            f"Average Latency: [value]{result.average_latency}[/value] [unit]{result.elapsed_time_unit.value}[/unit] | "
            f"Maximum Latency: [value]{result.maximum_latency}[/value] [unit]{result.elapsed_time_unit.value}[/unit] |"
            f"Median Deviation: [value]{result.median_deviation}[/value] [unit]{result.elapsed_time_unit.value}[/unit]\n"
            f"Packets Transmitted: [value]{result.packets_transmitted}[/value] | "
            f"Packets Received: [value]{result.packets_transmitted}[/value] | "
            f"Packets Lost: [value]{result.packets_lost}[/value] [unit]{result.packets_lost_unit.value}[/unit]\n"
            f"Elapsed Time: [value]{result.elapsed_time}[/value] [unit]{result.elapsed_time_unit.value}[/unit]"
        )
    console.rule()
    console.print(output)
    console.rule()
    return ExitStatus.success


@cli.command("netflix_fast")
def perform_netflix_fast_measurement():
    """
    Perform a Netflix fast.com measurement.
    """

    console = Console(theme=OUTPUT_THEME)
    try:
        measurement = NetflixFastMeasurement(
            id=get_uuid_str(),
        )
    except ValueError as err:
        raise click.BadParameter(err)
    with Halo(text="Performing Netflix Fast measurement", spinner="dots"):
        results = measurement.measure()
    # Extract host details from thread results
    output = "[header]:popcorn:   Netflix Fast   :popcorn:[/header]\n"
    for index, result in enumerate(
        [r for r in results if type(r) == NetflixFastThreadResult]
    ):
        if len(result.errors) > 0:
            for error in result.errors:
                console.print(f"[error]Error:[/error] {error.description}")
            return ExitStatus.failure
        output += (
            f"Host {index}: [value]{result.host}[/value] | "
            f"Country: [value]{result.country}[/value] | "
            f"City: [value]{result.city}[/value]\n"
        )
    for result in [r for r in results if type(r) == NetflixFastMeasurementResult]:
        if len(result.errors) > 0:
            for error in result.errors:
                console.print(f"[error]Error:[/error] {error.description}")
        output += (
            f"Download Rate: [value]{result.download_rate}[/value] [unit]{result.download_rate_unit.value}[/unit] | "
            f"Download Size: [value]{result.download_size}[/value] [unit]{result.download_size_unit.value}[/unit]"
        )
    console.rule()
    console.print(output)
    console.rule()
    return ExitStatus.success


@cli.command("speedtest_dotnet")
def perform_speedtest_dotnet_measurement():
    """
    Perform a speedtest.net measurement.
    """

    console = Console(theme=OUTPUT_THEME)
    try:
        measurement = SpeedtestDotnetMeasurement(
            id=get_uuid_str(),
        )
    except ValueError as err:
        raise click.BadParameter(err)
    with Halo(text="Performing Speedtest Dotnet measurement", spinner="dots"):
        result = measurement.measure()
    if len(result.errors) > 0:
        for error in result.errors:
            console.print(f"[error]Error:[/error] {error.description}")
            return ExitStatus.failure
    output = (
        f"[header]:zap: Speedtest Dotnet :zap:[/header]\n"
        f"Host: [endpoint]{result.server_host}[/endpoint] | "
        f"Name: [value]{result.server_name}[/value] | "
        f"ID: [value]{result.server_id}[/value] | "
        f"Sponsor: [value]{result.server_sponsor}[/value]\n"
        f"Download Rate: [value]{result.download_rate}[/value] [unit]{result.download_rate_unit.value}[/unit] | "
        f"Upload Rate: [value]{result.upload_rate}[/value] [unit]{result.upload_rate_unit.value}[/unit]"
    )
    console.rule()
    console.print(output)
    console.rule()
    return ExitStatus.success


@cli.command("webpage_download")
@click.option(
    "-u", "--url", required=True, multiple=False, help="URL of webpage to download"
)
def perform_webpage_download_measurement(url):
    """
    Perform a webpage download measurement.
    """

    console = Console(theme=OUTPUT_THEME)
    try:
        measurement = WebpageDownloadMeasurement(
            id=get_uuid_str(),
            url=url,
        )
    except ValueError as err:
        raise click.BadParameter(err)
    with Halo(text="Performing Webpage Download measurement", spinner="dots"):
        result = measurement.measure()
    if len(result.errors) > 0:
        for error in result.errors:
            console.print(f"[error]Error:[/error] {error.description}")
            return ExitStatus.failure
    output = (
        f"[header]:globe_with_meridians: Webpage Download :globe_with_meridians:[/header]\n"
        f"URL: [endpoint]{result.url}[/endpoint]\n"
        f"Download Rate: [value]{result.download_rate}[/value] [unit]{result.download_rate_unit.value}[/unit] | "
        f"Download Size: [value]{result.download_size}[/value] [unit]{result.download_size_unit.value}[/unit]\n"
        f"Elapsed Time: [value]{result.elapsed_time}[/value] [unit]{result.elapsed_time_unit.value}[/unit] | "
        f"Asset Count: [value]{result.asset_count}[/value] | "
        f"Failed Asset Downloads: [value]{result.failed_asset_downloads}[/value]"
    )
    console.rule()
    console.print(output)
    console.rule()
    return ExitStatus.success


@cli.command("youtube_download")
@click.option(
    "-u",
    "--url",
    required=True,
    multiple=False,
    help="URL of YouTube video to download",
)
def perform_youtube_download_measurement(url):
    """
    Perform a youtube download measurement.
    """
    console = Console(theme=OUTPUT_THEME)
    try:
        measurement = YoutubeDownloadMeasurement(
            id=get_uuid_str(),
            url=url,
        )
    except ValueError as err:
        raise click.BadParameter(err)
    with Halo(text="Performing Youtube Download measurement", spinner="dots"):
        result = measurement.measure()
    if len(result.errors) > 0:
        for error in result.errors:
            console.print(f"[error]Error:[/error] {error.description}")
    output = (
        f"[header]:tv: Youtube Download :tv:[/header]\n"
        f"URL: [endpoint]{result.url}[/endpoint]\n"
        f"Download Rate: [value]{result.download_rate}[/value] [unit]{result.download_rate_unit.value}[/unit] | "
        f"Download Size: [value]{result.download_size}[/value] [unit]{result.download_size_unit.value}[/unit]\n"
        f"Elapsed Time: [value]{result.elapsed_time}[/value] [unit]{result.elapsed_time_unit.value}[/unit]"
    )
    console.rule()
    console.print(output)
    console.rule()
    return ExitStatus.success



import typing
from dataclasses import dataclass

from netmeasure.measurements.base.results import MeasurementResult
from netmeasure.units import TimeUnit, StorageUnit, RatioUnit


@dataclass(frozen=True)
class LatencyMeasurementResult(MeasurementResult):
    """Encapsulates the results from a latency measurement.

    :param host: The host that was used to perform the latency
    measurement.
    :param minimum_latency: The minimum amount of latency witnessed
    while performing the measurement.
    :param average_latency: The average amount of latency witnessed
    while performing the measurement.
    :param maximum_latency: The maximum amount of latency witnessed
    while performing the measurement.
    :param median_deviation: The median deviation witnessed across
    the measurement.
    """

    host: str
    minimum_latency: typing.Optional[float]
    average_latency: typing.Optional[float]
    maximum_latency: typing.Optional[float]
    median_deviation: typing.Optional[float]
    packets_transmitted: typing.Optional[int]
    packets_received: typing.Optional[int]
    packets_lost: typing.Optional[float]
    packets_lost_unit: typing.Optional[RatioUnit]
    elapsed_time: typing.Optional[float]
    elapsed_time_unit: typing.Optional[TimeUnit]


@dataclass(frozen=True)
class LatencyIndividualMeasurementResult(MeasurementResult):
    host: str
    packet_size: typing.Optional[float]
    packet_size_unit: typing.Optional[StorageUnit]
    reverse_dns_address: typing.Optional[str]
    ip_address: typing.Optional[str]
    icmp_sequence: typing.Optional[int]
    time_to_live: typing.Optional[float]
    elapsed_time: typing.Optional[float]
    elapsed_time_unit: typing.Optional[TimeUnit]



from enum import Enum


class NetworkUnit(Enum):
    """Anticipated units for a network style of measurement."""

    bit_per_second = "bit/s"
    kilobit_per_second = "kbit/s"
    megabit_per_second = "Mbit/s"
    kibibit_per_second = "Kibit/s"
    mebibit_per_second = "Mibit/s"
    byte_per_second = "Byte/s"


class StorageUnit(Enum):
    """Anticipated units for a storage style of measurement."""

    bit = "b"
    byte = "B"
    kilobit = "kb"
    megabit = "Mb"
    kibibit = "Kib"
    mebibit = "Mib"
    kilobyte = "kB"
    megabyte = "MB"
    kibibyte = "KiB"
    mebibyte = "MiB"


class TimeUnit(Enum):
    millisecond = "ms"
    second = "s"
    minute = "m"
    hour = "h"
    day = "d"


class RatioUnit(Enum):
    percentage = "%"



import socket
import urllib
import time
import tempfile
import os
import shutil

import yt_dlp
import validators
from validators import ValidationFailure

from netmeasure.measurements.base.measurements import BaseMeasurement
from netmeasure.measurements.base.results import Error
from netmeasure.units import RatioUnit, TimeUnit, StorageUnit, NetworkUnit
from netmeasure.measurements.youtube_download.results import (
    YoutubeDownloadMeasurementResult,
)

YOUTUBE_ERRORS = {
    "youtube-download": "Download utility could not download file",
    "youtube-extractor": "Unable to extract info from youtube_download",
    "youtube-url": "Could not recognise URL",
    "youtube-attribute": "Could not parse attributes from progress dict",
    "youtube-progress_length": "Recorded progress dicts were too short",
    "youtube-file": "Could not remove file!!",
    "youtube-no_directory": "Could not find directory!!",
    "youtube-directory_nonempty": "Could not remove directory, non-empty!",
}


class YoutubeDownloadMeasurement(BaseMeasurement):
    def __init__(self, id, url):
        super(YoutubeDownloadMeasurement, self).__init__(id=id)
        validated_url = validators.url(url)
        if isinstance(validated_url, ValidationFailure):
            raise ValueError("`{url}` is not a valid url".format(url=url))
        self.id = id
        self.url = url
        self.progress_dicts = []

    def measure(self):
        return self._get_youtube_download_result(self.url)

    def _get_youtube_download_result(self, url):
        # Unique filename from process ID and timestamp
        file_dir = "{}/youtube-dl_{}".format(tempfile.gettempdir(), os.getpid())
        filename = "{}/youtube-dl_{}/{}".format(
            tempfile.gettempdir(), os.getpid(), int(time.time())
        )
        params = {
            "quiet": True,
            "no_progress": True,
            "progress_hooks": [self._store_progress_dicts_hook],
            "outtmpl": filename,
        }
        ydl = yt_dlp.YoutubeDL(params=params)
        try:
            ydl.extract_info(url)
        except yt_dlp.utils.ExtractorError as e:
            return self._get_youtube_download_error(
                "youtube-extractor", traceback=str(e)
            )
        except yt_dlp.utils.DownloadError as e:
            return self._get_youtube_download_error(
                "youtube-download", traceback=str(e)
            )
        try:
            # Extract size and duration from final progress step
            download_size = self.progress_dicts[-1]["total_bytes"]
            elapsed_time = self.progress_dicts[-1]["elapsed"]

            # Speed is only reported in non-final steps
            download_rate = self.progress_dicts[-2]["speed"] * 8
        except KeyError:
            return self._get_youtube_download_error(
                "youtube-attribute", traceback=str(self.progress_dicts)
            )
        except IndexError:
            return self._get_youtube_download_error(
                "youtube-progress_length", traceback=str(self.progress_dicts)
            )

        try:
            # Remove the created temp directory and all contents
            shutil.rmtree(file_dir)
        except FileNotFoundError as e:
            return self._get_youtube_download_error(
                "youtube-no_directory", traceback=str(e)
            )

        return YoutubeDownloadMeasurementResult(
            id=self.id,
            url=self.url,
            download_rate=download_rate,
            download_rate_unit=NetworkUnit("bit/s"),
            download_size=download_size,
            download_size_unit=StorageUnit("B"),
            elapsed_time=elapsed_time,
            elapsed_time_unit=TimeUnit("s"),
            errors=[],
        )

    def _store_progress_dicts_hook(self, s):
        """
        Saves the results of the download progress to a list for later parsing.
        This function is called at every progress step in the download utility
        """
        self.progress_dicts.append(s)

    def _get_youtube_download_error(self, key, traceback):
        return YoutubeDownloadMeasurementResult(
            id=self.id,
            url=self.url,
            download_rate_unit=None,
            download_rate=None,
            download_size=None,
            download_size_unit=None,
            elapsed_time=None,
            elapsed_time_unit=None,
            errors=[
                Error(
                    key=key,
                    description=YOUTUBE_ERRORS.get(key, ""),
                    traceback=traceback,
                )
            ],
        )



import re
import subprocess

import validators
from validators import ValidationFailure

from netmeasure.measurements.base.measurements import BaseMeasurement
from netmeasure.measurements.latency.results import (
    LatencyMeasurementResult,
    LatencyIndividualMeasurementResult,
)
from netmeasure.measurements.base.results import Error
from netmeasure.units import RatioUnit, TimeUnit, StorageUnit

LATENCY_OUTPUT_REGEX = re.compile(
    r"= (?P<minimum_latency>[\d.].*)/(?P<average_latency>[\d.].*)/(?P<maximum_latency>[\d.].*)/(?P<median_deviation>[\d.].*) "
)
LATENCY_PACKETS_REGEX = re.compile(
    r"(?P<packets_transmitted>\d*) packets transmitted, (?P<packets_received>\d*) received, "
    r"(?P<packet_loss>\d*)% packet loss, time (?P<time>\d*)(?P<time_unit>.*)"
)
LATENCY_INDIVIDUAL_PING_REGEX = re.compile(
    r"(?P<packet_size>\d*) (?P<packet_size_units>.*) from (?P<reverse_dns_address>.*) \(("
    r"?P<ip_address>.*)\): icmp_seq=(?P<icmp_sequence>\d*) ttl=(?P<time_to_live>\d*) time=("
    r"?P<time>.*) (?P<time_unit>.*)"
)


LATENCY_ERRORS = {
    "ping-err": "ping had an unknown error",
    "ping-split": "ping attempted to split the result but it was in an unanticipated format",
    "ping-regex": "ping attempted get the known regex format and failed.",
    "ping-minimum-latency": "ping could not process the minimum latency.",
    "ping-maximum-latency": "ping could not process the maximum latency.",
    "ping-average-latency": "ping could not process the average latency.",
    "ping-median-deviation": "ping could not process the median deviation.",
    "ping-no-server": "No closest server could be resolved.",
    "ping-timeout": "Measurement request timed out.",
}


class LatencyMeasurement(BaseMeasurement):
    def __init__(self, id, host, count=4, include_individual_results=False):
        super(LatencyMeasurement, self).__init__(id=id)
        if count < 1:
            raise ValueError(
                "A value of {count} was provided for the number of pings. This must be a positive "
                "integer greater than 0.".format(count=count)
            )

        validated_domain = validators.domain(host)
        validated_ip = validators.ipv4(host)
        if isinstance(validated_domain, ValidationFailure) & isinstance(
            validated_ip, ValidationFailure
        ):
            raise ValueError("`{host}` is not a valid host".format(host=host))

        self.host = host
        self.count = count
        self.include_individual_results = include_individual_results

    def measure(self):
        return self._get_latency_results(
            self.host,
            count=self.count,
            include_individual_results=self.include_individual_results,
        )

    def _get_latency_results(  # noqa: C901
        self, host, count=4, include_individual_results=False
    ):
        """Perform the latency measurement.

        :param host: The host name to perform the test against.
        :param count: The number of pings to determine latency with.
        :param include_individual_results: Should each of the
        individualised ping iterations be included in the results?
        :return: A list of `LatencyMeasurementResult` and
        `LatencyIndividualMeasurementResult` if individual results are
        enabled.
        """
        if host is None:
            return [self._get_latency_error("ping-no-server", host, traceback=None)]

        latency_out = subprocess.run(
            ["ping", "-c", "{c}".format(c=count), "{h}".format(h=host)],
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            universal_newlines=True,
        )

        # Note: only this error cares about stderr, other issues will be evident in stdout
        if latency_out.returncode != 0:
            return [
                self._get_latency_error("ping-err", host, traceback=latency_out.stderr)
            ]

        try:
            latency_data = latency_out.stdout.split("\n")[-2]
        except IndexError:
            return [
                self._get_latency_error(
                    "ping-split", host, traceback=latency_out.stdout
                )
            ]

        matches = LATENCY_OUTPUT_REGEX.search(latency_data)
        try:
            match_data = matches.groupdict()
        except AttributeError:
            return [
                self._get_latency_error(
                    "ping-regex", host, traceback=latency_out.stdout
                )
            ]

        if len(match_data.keys()) != 4:
            return [
                self._get_latency_error(
                    "ping-regex", host, traceback=latency_out.stdout
                )
            ]
        match_data = matches.groupdict()

        try:
            maximum_latency = float(match_data.get("maximum_latency"))
        except (TypeError, ValueError):
            return [
                self._get_latency_error(
                    "ping-maximum-latency", host, traceback=latency_out.stdout
                )
            ]

        try:
            minimum_latency = float(match_data.get("minimum_latency"))
        except (TypeError, ValueError):
            return [
                self._get_latency_error(
                    "ping-minimum-latency", host, traceback=latency_out.stdout
                )
            ]

        try:
            average_latency = float(match_data.get("average_latency"))
        except (TypeError, ValueError):
            return [
                self._get_latency_error(
                    "ping-average-latency", host, traceback=latency_out.stdout
                )
            ]

        try:
            median_deviation = float(match_data.get("median_deviation"))
        except (TypeError, ValueError):
            return [
                self._get_latency_error(
                    "ping-median_deviation", host, traceback=latency_out.stdout
                )
            ]

        try:
            latency_data = latency_out.stdout.split("\n")[-3]
        except IndexError:
            return [
                self._get_latency_error(
                    "ping-split", host, traceback=latency_out.stdout
                )
            ]

        matches = LATENCY_PACKETS_REGEX.search(latency_data)
        try:
            match_data = matches.groupdict()
        except AttributeError:
            return [
                self._get_latency_error(
                    "ping-regex", host, traceback=latency_out.stdout
                )
            ]

        packets_transmitted = int(match_data.get("packets_transmitted"))
        packets_received = int(match_data.get("packets_received"))
        packet_loss = float(match_data.get("packet_loss"))
        elapsed_time = float(match_data.get("time"))
        elapsed_time_unit = match_data.get("time_unit")

        results = [
            LatencyMeasurementResult(
                id=self.id,
                host=host,
                minimum_latency=minimum_latency,
                average_latency=average_latency,
                maximum_latency=maximum_latency,
                median_deviation=median_deviation,
                packets_transmitted=packets_transmitted,
                packets_received=packets_received,
                packets_lost=packet_loss,
                packets_lost_unit=RatioUnit.percentage,
                elapsed_time=elapsed_time,
                elapsed_time_unit=TimeUnit(elapsed_time_unit),
                errors=[],
            )
        ]

        if include_individual_results:
            matches = LATENCY_INDIVIDUAL_PING_REGEX.findall(latency_out.stdout)
            for match in matches:
                results.append(
                    LatencyIndividualMeasurementResult(
                        id=self.id,
                        host=host,
                        errors=[],
                        packet_size=match[0],
                        packet_size_unit=StorageUnit(match[1].replace("bytes", "B")),
                        reverse_dns_address=match[2],
                        ip_address=match[3],
                        icmp_sequence=match[4],
                        time_to_live=match[5],
                        elapsed_time=match[6],
                        elapsed_time_unit=TimeUnit(match[7]),
                    )
                )

        return results

    def _get_latency_error(self, key, host, traceback):
        return LatencyMeasurementResult(
            id=self.id,
            host=host,
            minimum_latency=None,
            average_latency=None,
            maximum_latency=None,
            median_deviation=None,
            packets_transmitted=None,
            packets_received=None,
            packets_lost=None,
            packets_lost_unit=None,
            elapsed_time=None,
            elapsed_time_unit=None,
            errors=[
                Error(
                    key=key,
                    description=LATENCY_ERRORS.get(key, ""),
                    traceback=traceback,
                )
            ],
        )



"""
Using the netflix_fast v2 api, the test collects some details about the client, and launches 1 thread per provided URL to download. Every `sleep_seconds` (presently 0.2) the test will append the latest speed (calculated by total downloaded bytes/total time taken) before checking for, in order:
    - `max_time_seconds` (presently 30s) expired.
    - Results have become "stabilised"
    - All threads have finished downloading
    - A single thread has finished downloading, IF `terminate_on_thread_complete=True`

Stabilisation is considered to be:
    - Downloaded has been running longer than `MIN_TIME_SECONDS` (presently 3s)
    - AND more than or equal to `MEASUREMENTS_COUNTED_BEFORE_CONSIDERED_STABLE` (presently 6) have been recorded
    - AND maximum percentage delta in these measurements is `< STABLE_MEASUREMENTS_DELTA` presently (2%)

In cases where the test concludes independently of the main loop (i.e when `reason_terminated == "thread_complete"`) The speed at the instant the thread completes is used, otherwise the final speed is used.

All this is then packaged into a `NetflixFastMeasurementResult`

After this, each of the URLs downloaded from has a latency test then have an Honesty-Box LatencyMeasurement test run against them, the results of which, along with the location and download rates/sizes for each thread are put into a `NetflixFastThreadResult`.

All these results are then returned as a list.
"""

import requests
import re
import time
import urllib
import json
from threading import Thread
from collections import deque
from statistics import mean

from netmeasure.measurements.base.measurements import BaseMeasurement
from netmeasure.measurements.base.results import Error
from netmeasure.units import RatioUnit, TimeUnit, StorageUnit, NetworkUnit
from netmeasure.measurements.latency.measurements import LatencyMeasurement
from netmeasure.measurements.file_download.measurements import FileDownloadMeasurement
from netmeasure.measurements.netflix_fast.results import (
    NetflixFastMeasurementResult,
    NetflixFastThreadResult,
)


NETFLIX_ERRORS = {
    "netflix-err": "Netflix test encountered an unknown error",
    "netflix-ping": "Netflix test encountered an error when pinging hosts",
    "netflix-response": "Netflix test received an invalid response from fast.com",
    "netflix-script-regex": "Netflix test failed to find script in the response",
    "netflix-script-response": "Netflix test received an invalid response from script",
    "netflix-token-regex": "Netflix test failed to find token in the response",
    "netflix-api-response": "Netflix test received an invalid response when querying for URLs",
    "netflix-api-json": "Netflix test failed to decode URLs",
    "netflix-api-parse": "Netflix test failed interpret elements of the decoded JSON",
    "netflix-connection": "Netflix test failed to connect to download URLs",
    "netflix-download": "Netflix test encountered an error downloading data",
}
MIN_TIME_SECONDS = 3
PING_COUNT = 4
MEASUREMENTS_COUNTED_BEFORE_CONSIDERED_STABLE = 6
STABLE_MEASUREMENTS_DELTA = 2
BITS_PER_BYTE = 8


class NetflixFastMeasurement(BaseMeasurement):
    def __init__(
        self,
        id,
        urlcount=3,
        max_time_seconds=30,
        sleep_seconds=0.2,
        chunk_size=64 * 2**10,
        terminate_on_thread_complete=True,
        terminate_on_result_stable=False,
    ):
        super(NetflixFastMeasurement, self).__init__(id=id)
        self.id = id
        self.urlcount = urlcount
        self.max_time_seconds = max_time_seconds
        self.sleep_seconds = sleep_seconds
        self.chunk_size = chunk_size
        self.terminate_on_thread_complete = terminate_on_thread_complete
        self.terminate_on_result_stable = terminate_on_result_stable
        self.finished_threads = 0
        self.exit_threads = False
        self.total = 0
        self.sessions = []
        self.client_data = {"asn": None, "ip": None, "isp": None, "location": None}
        self.targets = []
        self.thread_results = []
        self.completed_total = 0
        self.completed_elapsed_time = None

    def measure(self):
        results = []
        # Generate thread results dict structure
        for i in range(self.urlcount):
            self.thread_results.append(
                {
                    "index": i,
                    "elapsed_time": None,
                    "download_size": 0,
                    "download_rate": 0,
                    "url": None,
                    "location": None,
                }
            )

        results.append(self._get_fast_result())
        for thread_result in self.thread_results:
            results = results + self._get_url_result(thread_result)
        return results

    def _get_fast_result(self):
        s = requests.Session()
        try:
            resp = self._get_response(s)
        except ConnectionError as e:
            return self._get_netflix_error("netflix-response", traceback=str(e))

        try:
            script = re.search(r'<script src="(.*?)">', resp.text).group(1)
        except AttributeError:
            return self._get_netflix_error("netflix-script-regex", traceback=resp.text)

        try:
            script_resp = s.get("https://fast.com{script}".format(script=script))
        except ConnectionError as e:
            return self._get_netflix_error("netflix-script-response", traceback=str(e))

        try:
            token = re.search(r'token:"(.*?)"', script_resp.text).group(1)
        except AttributeError:
            return self._get_netflix_error(
                "netflix-token-regex", traceback=script_resp.text
            )

        try:
            self._query_api(s, token)
        except ConnectionError as e:
            return self._get_netflix_error("netflix-api-response", traceback=str(e))
        except json.decoder.JSONDecodeError as e:
            return self._get_netflix_error("netflix-api-json", traceback=str(e))
        except TypeError as e:
            return self._get_netflix_error("netflix-api-parse", traceback=str(e))
        except KeyError as e:
            return self._get_netflix_error("netflix-api-parse", traceback=str(e))

        try:
            conns = [
                self._get_connection(target["url"]) for target in self.thread_results
            ]
        except ConnectionError as e:
            return self._get_netflix_error("netflix-connection", traceback=str(e))

        fast_data = self._manage_threads(conns)

        return NetflixFastMeasurementResult(
            id=self.id,
            download_rate=float(fast_data["speed_bits"]),
            download_rate_unit=NetworkUnit("bit/s"),
            download_size=float(fast_data["total"]),
            download_size_unit=StorageUnit("B"),
            asn=self.client_data["asn"],
            ip=self.client_data["ip"],
            isp=self.client_data["isp"],
            city=self.client_data["location"]["city"],
            country=self.client_data["location"]["country"],
            urlcount=self.urlcount,
            reason_terminated=fast_data["reason_terminated"],
            errors=[],
        )

    def _manage_threads(self, conns):
        # Create worker threads
        threads = [None] * len(self.thread_results)
        for i in range(len(self.thread_results)):
            threads[i] = Thread(
                target=self._threaded_download,
                args=(conns[i], self.thread_results[i], time.time()),
            )
            threads[i].daemon = True
            threads[i].start()

        # Record approximate time
        start_time = time.time()
        recent_measurements = deque(
            maxlen=MEASUREMENTS_COUNTED_BEFORE_CONSIDERED_STABLE
        )
        recent_percent_deltas = deque(
            maxlen=MEASUREMENTS_COUNTED_BEFORE_CONSIDERED_STABLE
        )
        while True:
            elapsed_time = time.time() - start_time
            total = 0
            for thread_result in self.thread_results:
                total += thread_result["download_size"]
            speed_bits = total / elapsed_time * BITS_PER_BYTE
            recent_measurements.append(speed_bits)

            if (
                len(recent_measurements)
                == MEASUREMENTS_COUNTED_BEFORE_CONSIDERED_STABLE
            ):
                # Calculate percentage difference to the average of the last ten measurements
                # Note that there is no outlier detection here, all measurements are treated as-is.
                recent_percent_deltas.append(
                    (speed_bits - mean(recent_measurements)) / speed_bits * 100
                )

            if self._is_test_complete(elapsed_time, recent_percent_deltas):
                reason_terminated = self._is_test_complete(
                    elapsed_time, recent_percent_deltas
                )
                self.exit_threads = True
                for thread in threads:
                    thread.join()

                if (self.completed_elapsed_time is not None) & (
                    reason_terminated == "thread_complete"
                ):
                    # Record the speed at the time the thread finished downloading
                    speed_bits = (
                        self.completed_total
                        / self.completed_elapsed_time
                        * BITS_PER_BYTE
                    )
                else:
                    elapsed_time = time.time() - start_time
                    speed_bits = total / elapsed_time * BITS_PER_BYTE

                return {
                    "speed_bits": speed_bits,
                    "total": total,
                    "reason_terminated": reason_terminated,
                }
            time.sleep(self.sleep_seconds)

    def _threaded_download(self, conn, thread_result, start_time):
        # Iterate through the URL content
        g = conn.iter_content(chunk_size=self.chunk_size)
        for chunk in g:
            if self.exit_threads:
                break
            thread_result["download_size"] += len(chunk)

        completed_time = time.time()
        elapsed_time = completed_time - start_time

        # If this is the first thread to complete, record the time and total at this point
        if self.completed_elapsed_time is None:
            self.completed_elapsed_time = elapsed_time
            for global_thread_result in self.thread_results:
                self.completed_total += global_thread_result["download_size"]

        thread_result["download_rate"] = (
            thread_result["download_size"] / elapsed_time * BITS_PER_BYTE
        )
        thread_result["elapsed_time"] = elapsed_time
        self.finished_threads += 1

    def _query_api(self, s, token):
        params = {"https": "true", "token": token, "urlCount": self.urlcount}
        # '/v2/' path returns all location data about the servers
        api_resp = s.get("https://api.fast.com/netflix/speedtest/v2", params=params)
        api_json = api_resp.json()
        for i in range(len(api_json["targets"])):
            self.thread_results[i]["url"] = api_json["targets"][i]["url"]
            self.thread_results[i]["location"] = api_json["targets"][i]["location"]
        self.client_data = api_json["client"]
        return

    def _is_stabilised(self, recent_percent_deltas, elapsed_time):
        return (
            elapsed_time > MIN_TIME_SECONDS
            and len(recent_percent_deltas)
            >= MEASUREMENTS_COUNTED_BEFORE_CONSIDERED_STABLE
            and max(recent_percent_deltas) < STABLE_MEASUREMENTS_DELTA
        )

    def _get_response(self, s):
        return s.get("http://fast.com/")

    def _get_connection(self, url):
        s = requests.Session()
        self.sessions.append(s)
        conn = s.get(url, stream=True)
        return conn

    def _is_test_complete(self, elapsed_time, recent_percent_deltas):
        if elapsed_time > self.max_time_seconds:
            return "time_expired"
        if (self.terminate_on_result_stable) & (
            self._is_stabilised(recent_percent_deltas, elapsed_time)
        ):
            return "result_stabilised"
        if self.finished_threads == len(self.thread_results):
            return "all_complete"
        if (self.terminate_on_thread_complete) & (self.finished_threads >= 1):
            return "thread_complete"
        return False

    def _get_url_result(self, thread_result):
        host = urllib.parse.urlparse(thread_result["url"]).netloc
        city = thread_result["location"]["city"]
        country = thread_result["location"]["country"]
        LatencyResult = LatencyMeasurement(self.id, host, count=PING_COUNT).measure()[0]
        return [
            NetflixFastThreadResult(
                id=self.id,
                host=host,
                city=city,
                country=country,
                download_size=thread_result["download_size"],
                download_size_unit=StorageUnit("B"),
                download_rate=thread_result["download_rate"],
                download_rate_unit=NetworkUnit("bit/s"),
                elapsed_time=thread_result["elapsed_time"],
                elapsed_time_unit=TimeUnit("s"),
                errors=[],
            ),
            LatencyResult,
        ]

    def _get_netflix_error(self, key, traceback):
        return NetflixFastMeasurementResult(
            id=self.id,
            download_rate=None,
            download_rate_unit=None,
            download_size=None,
            download_size_unit=None,
            asn=None,
            ip=None,
            isp=None,
            city=None,
            country=None,
            urlcount=self.urlcount,
            reason_terminated=None,
            errors=[
                Error(
                    key=key,
                    description=NETFLIX_ERRORS.get(key, ""),
                    traceback=traceback,
                )
            ],
        )



import subprocess
from unittest import TestCase, mock

import speedtest

from netmeasure.measurements.speedtest_dotnet.measurements import (
    SpeedtestDotnetMeasurement,
    SPEEDTEST_ERRORS,
)
from netmeasure.measurements.speedtest_dotnet.results import (
    SpeedtestDotnetMeasurementResult,
)
from netmeasure.measurements.base.results import Error
from netmeasure.units import RatioUnit, TimeUnit, StorageUnit, NetworkUnit


class SpeedtestDotnetTestCase(TestCase):
    maxDiff = None

    def setUp(self) -> None:
        super().setUp()
        self.id = "1"
        self.stdnm = SpeedtestDotnetMeasurement("1")
        self.sample_results_dict_valid = {
            "download": 93116804.64881887,
            "upload": 19256654.06593738,
            "ping": 17.054,
            "server": {
                "url": "http://fake.site:8080/speedtest/upload.php",
                "lat": "-33.8600",
                "lon": "151.2111",
                "name": "HonestyVille",
                "country": "Australia",
                "cc": "AU",
                "sponsor": "'Yes' HonestyBox",
                "id": "1267",
                "url2": "http://s1.fake.site:8080/speedtest/upload.php",
                "host": "fake.site:8080",
                "d": 53.70823411720704,
                "latency": 17.054,
            },
            "timestamp": "2020-03-11T07:09:52.890803Z",
            "bytes_sent": 25591808,
            "bytes_received": 116746522,
            "share": "http://www.faketest.net/result/9117363621.png",
            "client": {
                "ip": "101.166.54.134",
                "lat": "-33.4102",
                "lon": "151.4225",
                "isp": "HonestyBox Internet",
                "isprating": "3.7",
                "rating": "0",
                "ispdlavg": "0",
                "ispulavg": "0",
                "loggedin": "0",
                "country": "AU",
            },
        }
        self.sample_result_valid = SpeedtestDotnetMeasurementResult(
            id=self.id,
            download_rate=93116804.64881887,
            download_rate_unit=NetworkUnit("bit/s"),
            upload_rate=19256654.06593738,
            upload_rate_unit=NetworkUnit("bit/s"),
            data_received=116746522,
            data_received_unit=StorageUnit("B"),
            latency=17.054,
            server_name="HonestyVille",
            server_id="1267",
            server_sponsor="'Yes' HonestyBox",
            server_host="fake.site:8080",
            errors=[],
        )
        self.sample_result_configretrieval = SpeedtestDotnetMeasurementResult(
            id=self.id,
            download_rate=None,
            download_rate_unit=None,
            upload_rate=None,
            upload_rate_unit=None,
            data_received=None,
            data_received_unit=None,
            latency=None,
            server_name=None,
            server_id=None,
            server_sponsor=None,
            server_host=None,
            errors=[
                Error(
                    key="speedtest-config",
                    description=SPEEDTEST_ERRORS.get("speedtest-config", ""),
                    traceback="<urlopen error [Errno -3] Temporary failure in name resolution>",
                )
            ],
        )
        self.sample_result_bestserver = SpeedtestDotnetMeasurementResult(
            id=self.id,
            download_rate=None,
            download_rate_unit=None,
            upload_rate=None,
            upload_rate_unit=None,
            data_received=None,
            data_received_unit=None,
            latency=None,
            server_name=None,
            server_id=None,
            server_sponsor=None,
            server_host=None,
            errors=[
                Error(
                    key="speedtest-best-server",
                    description=SPEEDTEST_ERRORS.get("speedtest-best-server", ""),
                    traceback="Unable to connect to servers to test latency.",
                )
            ],
        )
        self.sample_result_share = SpeedtestDotnetMeasurementResult(
            id=self.id,
            download_rate=None,
            download_rate_unit=None,
            upload_rate=None,
            upload_rate_unit=None,
            data_received=None,
            data_received_unit=None,
            latency=None,
            server_name=None,
            server_id=None,
            server_sponsor=None,
            server_host=None,
            errors=[
                Error(
                    key="speedtest-share",
                    description=SPEEDTEST_ERRORS.get("speedtest-share", ""),
                    traceback="<urlopen error [Errno -3] Temporary failure in name resolution>",
                )
            ],
        )

    @mock.patch("speedtest.Speedtest")
    def test_speedtest(self, mock_speedtest_constructor):
        run_mock = mock.Mock()
        results_mock = mock.Mock()
        results_mock.share = mock.Mock(return_value=None)
        results_mock.dict = mock.Mock(return_value=self.sample_results_dict_valid)
        run_mock.results = results_mock

        mock_speedtest_constructor.return_value = run_mock
        result = self.stdnm.measure()
        self.assertEqual(result, self.sample_result_valid)

    @mock.patch("speedtest.Speedtest")
    def test_speedtest_config_failure(self, mock_speedtest_constructor):
        mock_speedtest_constructor.side_effect = speedtest.ConfigRetrievalError(
            "<urlopen error [Errno -3] Temporary failure in name resolution>"
        )
        result = self.stdnm.measure()
        self.assertEqual(result, self.sample_result_configretrieval)

    @mock.patch("speedtest.Speedtest")
    def test_speedtest_get_best_server_failure(self, mock_speedtest_constructor):
        run_mock = mock.Mock()
        run_mock.get_best_server.side_effect = speedtest.SpeedtestBestServerFailure(
            "Unable to connect to servers to test latency."
        )
        mock_speedtest_constructor.return_value = run_mock
        result = self.stdnm.measure()
        self.assertEqual(result, self.sample_result_bestserver)

    @mock.patch("speedtest.Speedtest")
    def test_speedtest_share_results_failure(self, mock_speedtest_constructor):
        run_mock = mock.Mock()
        results_mock = mock.Mock()
        results_mock.share = mock.Mock(
            side_effect=speedtest.ShareResultsConnectFailure(
                "<urlopen error [Errno -3] Temporary failure in name resolution>"
            )
        )
        run_mock.results = results_mock
        mock_speedtest_constructor.return_value = run_mock
        result = self.stdnm.measure(share=True)
        self.assertEqual(result, self.sample_result_share)



"""
ip_route measurement, formerly the Traceroute measurement.
In a similar manner to the download_test, a list of hosts can be passed, in which case the one with the lowest latency is chosen.
The results are returned in order as:
 - IPRouteMeasurementResult
 - LatencyResult of the host used to create the IPRouteMeasurement
 - The LatencyResults from determining the initial latency
Traceroute measurement is provided by the "scapy" library. Note that this library needs route privileges to create
raw sockets, or to have the CAP_NET_RAW capability set.

"""
import socket
import validators
from validators import ValidationFailure

# Both of these imports seem to be required in order to mock the traceroute function
import scapy
from scapy.layers.inet import traceroute


from netmeasure.measurements.base.measurements import BaseMeasurement
from netmeasure.measurements.base.results import Error
from netmeasure.measurements.ip_route.results import IPRouteMeasurementResult
from netmeasure.measurements.latency.measurements import LatencyMeasurement

ROUTE_ERRORS = {
    "route-err": "iproute encountered an unknown error",
    "route-address": "iproute failed to find socket address",
    "route-permission": "iproute does not have permission to create a raw socket",
}


class IPRouteMeasurement(BaseMeasurement):
    def __init__(self, id, hosts, route_timeout=10, count=4):
        super(IPRouteMeasurement, self).__init__(id=id)

        if len(hosts) < 1:
            raise ValueError("At least one host must be provided.")
        for host in hosts:
            validated_domain = validators.domain(host)
            validated_ip = validators.ipv4(host)
            if isinstance(validated_domain, ValidationFailure) & isinstance(
                validated_ip, ValidationFailure
            ):
                raise ValueError("`{host}` is not a valid host".format(host=host))

        if count < 0:
            raise ValueError(
                "A value of {count} was provided for the number of pings. This must be a positive "
                "integer or `0` to turn off the ping.".format(count=count)
            )

        self.id = id
        self.hosts = hosts
        self.route_timeout = route_timeout
        self.count = count

    def measure(self):
        initial_latency_results = self._find_least_latent_host(self.hosts)
        least_latent_host = initial_latency_results[0][0]
        results = [self._get_traceroute_result(least_latent_host)]
        if self.count > 0:
            latency_measurement = LatencyMeasurement(
                self.id, least_latent_host, count=self.count
            )
            results.append(latency_measurement.measure()[0])
        results.extend([res for _, res in initial_latency_results])
        return results

    def _find_least_latent_host(self, hosts):
        """
        Performs a latency test for each specified host
        Returns a sorted list of LatencyResults, sorted by average latency
        """
        initial_latency_results = []
        for host in hosts:
            latency_measurement = LatencyMeasurement(self.id, host, count=2)
            initial_latency_results.append((host, latency_measurement.measure()[0]))
        return sorted(
            initial_latency_results,
            key=lambda x: (x[1].average_latency is None, x[1].average_latency),
        )

    def _get_traceroute_result(self, host):
        try:
            # Test whether raw socket privileges exist:
            socket.socket(socket.AF_PACKET, socket.SOCK_RAW)

            # Commence traceroute test:
            traceroute_out = scapy.layers.inet.traceroute(host, verbose=0)
            traceroute_trace = traceroute_out[0].get_trace()
            ip = list(traceroute_trace.keys())[0]
            hop_count = len(traceroute_trace[ip])
            trace_list = [
                x[0] for x in traceroute_trace[ip].values()
            ]  # Reshape dict into list of ips
        except socket.gaierror as e:
            return self._get_ip_route_error("route-address", traceback=str(e))
        except PermissionError as e:
            return self._get_ip_route_error("route-permission", traceback=str(e))

        return IPRouteMeasurementResult(
            id=self.id,
            host=host,
            ip=ip,
            hop_count=hop_count,
            route=trace_list,
            errors=[],
        )

    def _get_ip_route_error(self, key, traceback):
        return IPRouteMeasurementResult(
            id=self.id,
            host=None,
            ip=None,
            hop_count=None,
            route=None,
            errors=[
                Error(
                    key=key, description=ROUTE_ERRORS.get(key, ""), traceback=traceback
                )
            ],
        )



import typing
from dataclasses import dataclass

from netmeasure.measurements.base.results import MeasurementResult
from netmeasure.units import NetworkUnit, StorageUnit


@dataclass(frozen=True)
class FileDownloadMeasurementResult(MeasurementResult):
    """Encapsulates the results from a download speed measurement.

    :param url: The URL that was used to perform the download speed
    measurement.
    :param download_size: The size of the download (excluding units)
    that was used to perform the download speed measurement.
    :param download_size_unit: The unit of measurement used
    to describe the `download_size`.
    :param download_rate: The rate measured in the download speed
    measurement excluding units:
    :param download_rate_unit: The unit of measurement used to
    measure the `download_rate`.
    """

    url: str
    download_size: typing.Optional[float]
    download_size_unit: typing.Optional[StorageUnit]
    download_rate: typing.Optional[float]
    download_rate_unit: typing.Optional[NetworkUnit]



import re
import subprocess

import validators
from validators import ValidationFailure
import speedtest

from netmeasure.measurements.base.measurements import BaseMeasurement
from netmeasure.measurements.speedtest_dotnet.results import (
    SpeedtestDotnetMeasurementResult,
)
from netmeasure.measurements.base.results import Error
from netmeasure.units import RatioUnit, TimeUnit, StorageUnit, NetworkUnit

SPEEDTEST_ERRORS = {
    "speedtest-err": "wget had an unknown error.",
    "speedtest-config": "speedtest failed to retrieve a config",
    "speedtest-best-server": "speedtest could not find the best server",
    "speedtest-share": "speedtest could not share results",
    "speedtest-convert": "could not convert result values",
}


class SpeedtestDotnetMeasurement(BaseMeasurement):
    def __init__(self, id, servers=None):
        super(SpeedtestDotnetMeasurement, self).__init__(id=id)
        self.id = id
        self.servers = servers

    def measure(self, share=False):
        """
        @params share: Boolean determining whether to generate a PNG on speedtest.net displaying the result of the test.
        """
        try:
            s = speedtest.Speedtest()
        except speedtest.ConfigRetrievalError as e:
            return self._get_speedtest_error("speedtest-config", traceback=str(e))

        s.get_servers(self.servers)

        try:
            s.get_best_server()
        except speedtest.SpeedtestBestServerFailure as e:
            return self._get_speedtest_error("speedtest-best-server", traceback=str(e))

        s.download()
        s.upload()

        if share:
            try:
                s.results.share()
            except speedtest.ShareResultsConnectFailure as e:
                return self._get_speedtest_error("speedtest-share", traceback=str(e))

        results_dict = s.results.dict()
        try:
            return SpeedtestDotnetMeasurementResult(
                id=self.id,
                download_rate=float(results_dict["download"]),
                download_rate_unit=NetworkUnit("bit/s"),
                upload_rate=float(results_dict["upload"]),
                upload_rate_unit=NetworkUnit("bit/s"),
                data_received=(results_dict["bytes_received"]),
                data_received_unit=StorageUnit("B"),
                latency=float(results_dict["ping"]),
                server_name=results_dict["server"]["name"],
                server_id=results_dict["server"]["id"],
                server_sponsor=results_dict["server"]["sponsor"],
                server_host=results_dict["server"]["host"],
                errors=[],
            )
        except ValueError as e:
            return self._get_speedtest_error("speedtest-convert", traceback=str(e))

    def _get_speedtest_error(self, key, traceback):
        return SpeedtestDotnetMeasurementResult(
            id=self.id,
            download_rate=None,
            download_rate_unit=None,
            upload_rate=None,
            upload_rate_unit=None,
            data_received=None,
            data_received_unit=None,
            latency=None,
            server_name=None,
            server_id=None,
            server_sponsor=None,
            server_host=None,
            errors=[
                Error(
                    key=key,
                    description=SPEEDTEST_ERRORS.get(key, ""),
                    traceback=traceback,
                )
            ],
        )



import socket
from unittest import TestCase, mock

from netmeasure.measurements.ip_route.measurements import (
    IPRouteMeasurement,
    ROUTE_ERRORS,
)
from netmeasure.measurements.latency.measurements import LatencyMeasurement
from netmeasure.measurements.ip_route.results import IPRouteMeasurementResult
from netmeasure.measurements.latency.results import LatencyMeasurementResult
from netmeasure.measurements.base.results import Error
from netmeasure.units import RatioUnit, TimeUnit, StorageUnit, NetworkUnit


class IPRouteTestCase(TestCase):
    maxDiff = None

    def setUp(self) -> None:
        super().setUp()
        self.id = "1"
        self.example_hosts_one = ["www.fakesitetwo.com"]
        self.example_hosts_three = [
            "www.fakesiteone.com",
            "www.fakesitetwo.com",
            "www.fakesitethree.com",
        ]
        self.iprm = IPRouteMeasurement(self.id, hosts=self.example_hosts_one, count=4)
        self.example_trace_five = {
            "final.ip": {
                1: ("first.ip", False),
                2: ("second.ip", False),
                3: ("third.ip", False),
                4: ("fourth.ip", False),
                5: ("final.ip", True),
            }
        }
        self.example_trace_five_list = [
            "first.ip",
            "second.ip",
            "third.ip",
            "fourth.ip",
            "final.ip",
        ]
        self.example_result_five = IPRouteMeasurementResult(
            id=self.id,
            host=self.example_hosts_one[0],
            hop_count=5,
            ip="final.ip",
            route=self.example_trace_five_list,
            errors=[],
        )
        self.example_result_permission_err = IPRouteMeasurementResult(
            id=self.id,
            host=None,
            hop_count=None,
            ip=None,
            route=None,
            errors=[
                Error(
                    key="route-permission",
                    description=ROUTE_ERRORS.get("route-permission", ""),
                    traceback="[Errno 1] Operation not permitted",
                )
            ],
        )
        self.example_result_address_err = IPRouteMeasurementResult(
            id=self.id,
            host=None,
            hop_count=None,
            ip=None,
            route=None,
            errors=[
                Error(
                    key="route-address",
                    description=ROUTE_ERRORS.get("route-address", ""),
                    traceback="[Errno -2] Name or service not known",
                )
            ],
        )
        self.example_latency_results_three = [
            (
                LatencyMeasurementResult(
                    id=self.id,
                    host="www.fakesiteone.com",
                    minimum_latency=None,
                    average_latency=None,
                    maximum_latency=None,
                    median_deviation=None,
                    errors=[],
                    packets_transmitted=None,
                    packets_received=None,
                    packets_lost=None,
                    packets_lost_unit=None,
                    elapsed_time=None,
                    elapsed_time_unit=None,
                ),
            ),
            (
                LatencyMeasurementResult(
                    id=self.id,
                    host="www.fakesitetwo.com",
                    minimum_latency=None,
                    average_latency=25.0,
                    maximum_latency=None,
                    median_deviation=None,
                    errors=[],
                    packets_transmitted=None,
                    packets_received=None,
                    packets_lost=None,
                    packets_lost_unit=None,
                    elapsed_time=None,
                    elapsed_time_unit=None,
                ),
            ),
            (
                LatencyMeasurementResult(
                    id=self.id,
                    host="www.fakesitethree.com",
                    minimum_latency=None,
                    average_latency=999.0,
                    maximum_latency=None,
                    median_deviation=None,
                    errors=[],
                    packets_transmitted=None,
                    packets_received=None,
                    packets_lost=None,
                    packets_lost_unit=None,
                    elapsed_time=None,
                    elapsed_time_unit=None,
                ),
            ),
        ]
        self.example_least_latent_result = (
            (
                LatencyMeasurementResult(
                    id=self.id,
                    host="www.fakesitetwo.com",
                    minimum_latency=None,
                    average_latency=24.9,
                    maximum_latency=None,
                    median_deviation=None,
                    errors=[],
                    packets_transmitted=None,
                    packets_received=None,
                    packets_lost=None,
                    packets_lost_unit=None,
                    elapsed_time=None,
                    elapsed_time_unit=None,
                ),
            ),
        )

    @mock.patch.object(socket, "socket")
    @mock.patch.object(LatencyMeasurement, "measure")
    @mock.patch("scapy.layers.inet.traceroute")
    def test_measure(self, mock_get_traceroute, mock_latency_results, mock_socket):
        iprm_three = IPRouteMeasurement(
            self.id, hosts=self.example_hosts_three, count=4
        )
        mock_trace = mock.MagicMock()
        mock_trace.get_trace.return_value = self.example_trace_five
        mock_get_traceroute.return_value = [mock_trace, None]
        mock_latency_results.side_effect = [
            self.example_latency_results_three[0],
            self.example_latency_results_three[1],
            self.example_latency_results_three[2],
            self.example_least_latent_result,
        ]
        self.assertEqual(
            iprm_three.measure(),
            [
                self.example_result_five,
                self.example_least_latent_result[0],
                self.example_latency_results_three[1][0],
                self.example_latency_results_three[2][0],
                self.example_latency_results_three[0][0],
            ],
        )

    @mock.patch.object(socket, "socket")
    @mock.patch("scapy.layers.inet.traceroute")
    def test_get_trace_five(self, mock_get_traceroute, mock_socket):
        mock_trace = mock.MagicMock()
        mock_trace.get_trace.return_value = self.example_trace_five
        mock_get_traceroute.return_value = [mock_trace, None]
        self.assertEqual(
            self.iprm._get_traceroute_result(self.example_hosts_one[0]),
            self.example_result_five,
        )

    @mock.patch.object(socket, "socket")
    @mock.patch("scapy.layers.inet.traceroute")
    def test_get_trace_permission_err(self, mock_get_traceroute, mock_socket):
        mock_get_traceroute.side_effect = PermissionError(
            "[Errno 1] Operation not permitted"
        )
        self.assertEqual(
            self.iprm._get_traceroute_result(self.example_hosts_one[0]),
            self.example_result_permission_err,
        )

    @mock.patch.object(socket, "socket")
    @mock.patch("scapy.layers.inet.traceroute")
    def test_get_trace_address_err(self, mock_get_traceroute, mock_socket):
        mock_get_traceroute.side_effect = socket.gaierror(
            "[Errno -2] Name or service not known"
        )
        self.assertEqual(
            self.iprm._get_traceroute_result(self.example_hosts_one[0]),
            self.example_result_address_err,
        )


class IPRouteMeasurementLeastLatentTestCase(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.id = "1"
        self.example_hosts_three = [
            "www.fakesiteone.com",
            "www.fakesitetwo.com",
            "www.fakesitethree.com",
        ]
        self.example_results_three = [
            (
                LatencyMeasurementResult(
                    id=self.id,
                    host="www.fakesiteone.com",
                    minimum_latency=None,
                    average_latency=None,
                    maximum_latency=None,
                    median_deviation=None,
                    errors=[],
                    packets_transmitted=None,
                    packets_received=None,
                    packets_lost=None,
                    packets_lost_unit=None,
                    elapsed_time=None,
                    elapsed_time_unit=None,
                ),
            ),
            (
                LatencyMeasurementResult(
                    id=self.id,
                    host="www.fakesitetwo.com",
                    minimum_latency=None,
                    average_latency=25.0,
                    maximum_latency=None,
                    median_deviation=None,
                    errors=[],
                    packets_transmitted=None,
                    packets_received=None,
                    packets_lost=None,
                    packets_lost_unit=None,
                    elapsed_time=None,
                    elapsed_time_unit=None,
                ),
            ),
            (
                LatencyMeasurementResult(
                    id=self.id,
                    host="www.fakesitethree.com",
                    minimum_latency=None,
                    average_latency=999.0,
                    maximum_latency=None,
                    median_deviation=None,
                    errors=[],
                    packets_transmitted=None,
                    packets_received=None,
                    packets_lost=None,
                    packets_lost_unit=None,
                    elapsed_time=None,
                    elapsed_time_unit=None,
                ),
            ),
        ]
        self.example_hosts_one = ["www.fakesiteone.com"]
        self.example_results_one = [
            LatencyMeasurementResult(
                id=self.id,
                host="www.fakesiteone.com",
                minimum_latency=None,
                average_latency=None,
                maximum_latency=None,
                median_deviation=None,
                errors=[],
                packets_transmitted=None,
                packets_received=None,
                packets_lost=None,
                packets_lost_unit=None,
                elapsed_time=None,
                elapsed_time_unit=None,
            ),
        ]
        self.iprm_three = IPRouteMeasurement(
            self.id, hosts=self.example_hosts_three, count=4
        )
        self.iprm_one = IPRouteMeasurement(
            self.id, hosts=self.example_hosts_three, count=4
        )

    @mock.patch.object(LatencyMeasurement, "measure")
    def test_sort_least_latent_host(self, mock_latency_results):
        mock_latency_results.side_effect = self.example_results_three
        self.assertEqual(
            self.iprm_three._find_least_latent_host(self.example_hosts_three),
            [
                (self.example_hosts_three[1], self.example_results_three[1][0]),
                (self.example_hosts_three[2], self.example_results_three[2][0]),
                (self.example_hosts_three[0], self.example_results_three[0][0]),
            ],
        )

    @mock.patch.object(LatencyMeasurement, "measure")
    def test_sort_one_host(self, mock_latency_results):
        mock_latency_results.return_value = self.example_results_one
        self.assertEqual(
            self.iprm_one._find_least_latent_host(self.example_hosts_one),
            [(self.example_hosts_one[0], self.example_results_one[0])],
        )


class IPRouteMeasurementCreationTestCase(TestCase):
    def test_invalid_hosts(self, *args):
        self.assertRaises(ValueError, IPRouteMeasurement, "test", ["invalid..host"])

    def test_invalid_count(self, *args):
        self.assertRaises(
            TypeError,
            IPRouteMeasurement,
            "test",
            ["validfakeurl.com"],
            count="invalid-count",
        )

    def test_invalid_negative_count(self, *args):
        self.assertRaises(
            ValueError, IPRouteMeasurement, "test", ["validfakeurl.com"], count=-2
        )

    def test_valid_ip_host(self):
        IPRouteMeasurement("test", ["1.1.1.1"])



import typing
from dataclasses import dataclass

from netmeasure.measurements.base.results import MeasurementResult
from netmeasure.units import TimeUnit, StorageUnit, RatioUnit, NetworkUnit


@dataclass(frozen=True)
class NetflixFastMeasurementResult(MeasurementResult):
    """Encapsulates the results from a NetflixFast measurement."""

    download_rate: typing.Optional[float]
    download_rate_unit: typing.Optional[NetworkUnit]
    download_size: typing.Optional[float]
    download_size_unit: typing.Optional[StorageUnit]
    asn: typing.Optional[str]
    ip: typing.Optional[str]
    isp: typing.Optional[str]
    city: typing.Optional[str]
    country: typing.Optional[str]
    urlcount: typing.Optional[int]
    reason_terminated: typing.Optional[str]


@dataclass(frozen=True)
class NetflixFastThreadResult(MeasurementResult):
    """Encapsulates the latency test results from an individual download url."""

    host: str
    download_size: typing.Optional[float]
    download_size_unit: typing.Optional[StorageUnit]
    download_rate: typing.Optional[float]
    download_rate_unit: typing.Optional[NetworkUnit]
    elapsed_time: typing.Optional[float]
    elapsed_time_unit: typing.Optional[TimeUnit]
    city: typing.Optional[str]
    country: typing.Optional[str]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['netmeasure',
 'netmeasure.measurements',
 'netmeasure.measurements.base',
 'netmeasure.measurements.file_download',
 'netmeasure.measurements.file_download.tests',
 'netmeasure.measurements.ip_route',
 'netmeasure.measurements.ip_route.tests',
 'netmeasure.measurements.latency',
 'netmeasure.measurements.latency.tests',
 'netmeasure.measurements.netflix_fast',
 'netmeasure.measurements.netflix_fast.tests',
 'netmeasure.measurements.speedtest_dotnet',
 'netmeasure.measurements.speedtest_dotnet.tests',
 'netmeasure.measurements.webpage_download',
 'netmeasure.measurements.webpage_download.tests',
 'netmeasure.measurements.youtube_download',
 'netmeasure.measurements.youtube_download.tests']

package_data = \
{'': ['*']}

install_requires = \
['beautifulsoup4>=4.11.2,<5.0.0',
 'click>=8.1.3,<9.0.0',
 'exitstatus>=2.3.0,<3.0.0',
 'halo>=0.0.31,<0.0.32',
 'requests>=2.28,<3.0',
 'rich>=13.3.2,<14.0.0',
 'scapy>=2.5,<3.0',
 'speedtest-cli>=2.1,<3.0',
 'validators>=0.20,<0.21',
 'yt-dlp>=2023.3.4,<2024.0.0']

entry_points = \
{'console_scripts': ['netmeasure = netmeasure.cli:cli']}

setup_kwargs = {
    'name': 'netmeasure',
    'version': '1.2.0',
    'description': 'A tool for measuring Internet connection quality in a structured way.',
    'long_description': '[![PyPI version](https://badge.fury.io/py/honestybox-measurement.svg)](https://badge.fury.io/py/honestybox-measurement)\n[![PyPI Supported Python Versions](https://img.shields.io/pypi/pyversions/honestybox-measurement.svg)](https://pypi.python.org/pypi/honestybox-measurement/)\n[![GitHub license](https://img.shields.io/github/license/honesty-box/honestybox-measurement)](https://github.com/honesty-box/honestybox-measurement/blob/master/LICENSE)\n[![GitHub Actions (Tests)](https://github.com/amorphitec/netmeasure/workflows/Tests/badge.svg)](https://github.com/amorphitec/netmeasure)\n\n# Netmeasure\n\nA library for measuring Internet connection quality in a structured and consistent way.\n\n## Purpose\n\nThere are a variety of services, clients, tools, and methodologies used to measure Internet connection quality. Each of these has its own advantages, flaws, biases and units of measurement.\n\nNetmeasure brings together a variety of Internet connection quality measurements in a single library with a consistent interface and explicitly-defined units.\n\nAn open-source license ensures methodology is transparent and open to ongoing community improvement.\n\n## Measurements\n\n- `file_download` - measures download of a file from a given endpoint using the [wget](https://www.gnu.org/software/wget/) application.\n- `ip_route` - measures network hops to a given endpoint using the [scapy](https://scapy.net/) library.\n- `latency` - measures latency to a given endpoint using the [ping](https://en.wikipedia.org/wiki/Ping_%28networking_utility%29) application.\n- `netflix_fast` - measures download from the [netflix fast](https://fast.com/) service using the [requests](https://requests.readthedocs.io/en/latest/) library.\n- `speedtest_dotnet` - measures download from, upload to and latency to the [speedtest.net](https://www.speedtest.net/) service using the [speedtest-cli](https://pypi.org/project/speedtest-cli/) library.\n- `webpage_download` - measures download of a given web page and its associated assets using the [requests](https://requests.readthedocs.io/en/latest/) library.\n- `youtube_download` - measures download of a given [youtube](https://www.youtube.com/) video using the [youtube-dl](https://youtube-dl.org/) library.\n\n## Usage\n\n...\n\n## Requirements\n\n`netmeasure` supports Python 3.8 to Python 3.11 inclusively.\n\n## Development\n\n### Git hooks\n\n[pre-commit](https://pre-commit.com/) hooks are included to ensure code quality\non `commit` and `push`. Install these hooks like so:\n\n```shell script\n$ pre-commit install && pre-commit install -t pre-push\nasd\n```\n\n### Publishing a release\n\n1. Install [poetry](https://poetry.eustace.io)\n\n2. Checkout the release:\n\n    ```shell script\n    $ git checkout v<x>.<y>.<z>\n    ```\n\n3. Publish the release:\n\n    ```shell script\n    $ poetry publish --build\n    ```\n',
    'author': 'James Stewart',
    'author_email': 'james@amorphitec.io',
    'maintainer': 'James Stewart',
    'maintainer_email': 'james@amorphitec.io',
    'url': 'https://github.com/amorphitec/netmeasure',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4',
}


setup(**setup_kwargs)



import time
from six.moves.urllib.parse import urlparse

import requests
from bs4 import BeautifulSoup

from netmeasure.measurements.base.measurements import BaseMeasurement
from netmeasure.measurements.base.results import Error
from netmeasure.units import RatioUnit, TimeUnit, StorageUnit, NetworkUnit
from netmeasure.measurements.webpage_download.results import (
    WebpageDownloadMeasurementResult,
)
from netmeasure.measurements.latency.measurements import LatencyMeasurement


WEB_ERRORS = {
    "web-get": "Failed to complete the initial connection",
    "web-parse": "Failed to parse assets from HTML",
    "web-parse-rel": "Failed to determine 'rel' attribute of link",
    "web-assets": "Failed to download secondary assets",
    "web-timeout": "Initial page download timed out",
}
VALID_LINK_EXTENSTIONS = [".css", ".ico", ".png", ".woff2", ""]
VALID_LINK_REL_ATTRIBUTES = [
    "manifest",
    "modulepreload",
    "preload",
    "prerender",
    "stylesheet",
    "apple-touch-icon",
    "icon",
    "shortcut icon",
]


class WebpageDownloadMeasurement(BaseMeasurement):
    def __init__(self, id, url, count=4, download_timeout=180):
        self.id = id
        self.url = url
        self.count = count
        self.download_timeout = download_timeout

    def measure(self):
        host = urlparse(self.url).netloc
        protocol = urlparse(self.url).scheme
        return self._get_webpage_result(self.url, host, protocol)

    def _get_webpage_result(self, url, host, protocol):
        s = requests.Session()
        headers = {
            "dnt": "1",
            "upgrade-insecure-requests": "1",
            "user-agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_4) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/83.0.4103.61 Safari/537.36",
            "accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9",
            "sec-fetch-site": "none",
            "sec-fetch-mode": "navigate",
            "sec-fetch-user": "?1",
            "sec-fetch-dest": "document",
            "accept-language": "en-GB,en-US;q=0.9,en;q=0.8",
        }

        start_time = time.time()
        try:
            r = s.get(url, headers=headers, timeout=self.download_timeout)
        except (ConnectionError, requests.ConnectionError) as e:
            return self._get_webpage_error("web-get", traceback=str(e))
        except requests.exceptions.ReadTimeout as e:
            return self._get_webpage_error("web-timeout", traceback=str(e))
        try:
            to_download = self._parse_html(r.text)
        except TypeError as e:
            return self._get_webpage_error("web-parse-rel", traceback=str(e))
        try:
            asset_download_metrics = self._download_assets(
                s, to_download, host, protocol
            )
        except TypeError as e:
            return self._get_webpage_error("web-assets", traceback=str(e))

        primary_download_size = len(r.text)
        asset_download_size = asset_download_metrics["asset_download_size"]
        elapsed_time = asset_download_metrics["completion_time"] - start_time
        download_rate = (primary_download_size + asset_download_size) * 8 / elapsed_time
        failed_asset_downloads = asset_download_metrics["failed_asset_downloads"]

        return WebpageDownloadMeasurementResult(
            id=self.id,
            url=url,
            download_rate=download_rate,
            download_rate_unit=NetworkUnit("bit/s"),
            download_size=primary_download_size + asset_download_size,
            download_size_unit=StorageUnit("B"),
            asset_count=len(to_download),
            failed_asset_downloads=failed_asset_downloads,
            elapsed_time=elapsed_time,
            elapsed_time_unit=TimeUnit("s"),
            errors=[],
        )

    def _parse_html(self, content):
        soup = BeautifulSoup(content, "html.parser")
        imgs = soup.find_all("img")
        links = soup.find_all("link")
        scripts = soup.find_all("script")
        to_download = []
        for img in imgs:
            if img.has_attr("src"):
                to_download.append(img["src"])
        for script in scripts:
            if script.has_attr("src"):
                to_download.append(script["src"])
        for link in links:
            if link.has_attr("rel") & link.has_attr("href"):
                # Join in the case where `rel` is more than one word
                rel = " ".join(link["rel"])
                if rel in VALID_LINK_REL_ATTRIBUTES:
                    to_download.append(link["href"])

        return to_download

    def _download_assets(self, session, to_download, host, protocol):
        # Store the amount of bytes downloaded
        asset_download_sizes = []
        failed_asset_downloads = 0
        for asset in to_download:
            try:
                # Identify data URLs (already downloaded inline, counted in main download size)
                if "data:" in asset:
                    continue

                # Check if path w/o preceeding slashes is a valid URL
                if asset.startswith("//"):
                    download_url = protocol + ":" + asset
                # Check if path is a relative path
                elif asset.startswith("/"):
                    download_url = protocol + "://" + host + asset
                # ...or simply a normal file link
                else:
                    download_url = asset

                a = session.get(download_url, timeout=self.download_timeout)
                if a.status_code >= 400:
                    raise ConnectionError
                asset_download_sizes.append(len(a.text))
            except ConnectionError:
                failed_asset_downloads = failed_asset_downloads + 1
            except requests.exceptions.MissingSchema:
                failed_asset_downloads = failed_asset_downloads + 1
            except requests.exceptions.ReadTimeout:
                failed_asset_downloads = failed_asset_downloads + 1

        return {
            "asset_download_size": sum(asset_download_sizes),
            "failed_asset_downloads": failed_asset_downloads,
            "completion_time": time.time(),
        }

    def _get_webpage_error(self, key, traceback):
        return WebpageDownloadMeasurementResult(
            id=self.id,
            url=self.url,
            download_rate_unit=None,
            download_rate=None,
            download_size=None,
            download_size_unit=None,
            asset_count=None,
            failed_asset_downloads=None,
            elapsed_time=None,
            elapsed_time_unit=None,
            errors=[
                Error(key=key, description=WEB_ERRORS.get(key, ""), traceback=traceback)
            ],
        )



import subprocess
import sys
import json
from unittest import TestCase, mock
from threading import active_count
from itertools import cycle

from netmeasure.measurements.netflix_fast.measurements import (
    NetflixFastMeasurement,
    NETFLIX_ERRORS,
    MIN_TIME_SECONDS,
    PING_COUNT,
    MEASUREMENTS_COUNTED_BEFORE_CONSIDERED_STABLE,
    STABLE_MEASUREMENTS_DELTA,
)
from netmeasure.measurements.netflix_fast.results import (
    NetflixFastMeasurementResult,
    NetflixFastThreadResult,
)
from netmeasure.measurements.latency.results import LatencyMeasurementResult
from netmeasure.measurements.base.results import Error
from netmeasure.units import RatioUnit, TimeUnit, StorageUnit, NetworkUnit


"""
Note that calling .measure() and ._get_fast_result() mutate the NetflixFastMeasurement object.
As a result these objects should preferably be created inside the test function if these functions are to be called.
"""


class NetflixResultTestCase(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.nft = NetflixFastMeasurement(
            "1", urlcount=3, terminate_on_thread_complete=True
        )
        self.api_response_three = {
            "client": {
                "location": {"city": "HBox City", "country": "HBoxtopia"},
                "isp": "ServiceProvider",
                "asn": "0000",
                "ip": "0000:0000:000a:0000:0c00:0b00:a0a0:f0f0",
            },
            "targets": [
                {
                    "location": {
                        "city": "Foreign City One",
                        "country": "Foreign Country One",
                    },
                    "name": "https://afakeurl.1.notreal.net/speedtest",
                    "url": "https://afakeurl.1.notreal.net/speedtest",
                },
                {
                    "location": {
                        "city": "Foreign City Two",
                        "country": "Foreign Country Two",
                    },
                    "name": "https://afakeurl.2.notreal.net/speedtest",
                    "url": "https://afakeurl.2.notreal.net/speedtest",
                },
                {
                    "location": {
                        "city": "Foreign City Three",
                        "country": "Foreign Country Three",
                    },
                    "name": "https://afakeurl.3.notreal.net/speedtest",
                    "url": "https://afakeurl.3.notreal.net/speedtest",
                },
            ],
        }
        self.fast_data_three = {
            "speed_bits": 1234,
            "total": 4321,
            "reason_terminated": "fake_reason",
        }
        self.fast_result_three = NetflixFastMeasurementResult(
            id="1",
            download_rate=float(self.fast_data_three["speed_bits"]),
            download_rate_unit=NetworkUnit("bit/s"),
            download_size=float(self.fast_data_three["total"]),
            download_size_unit=StorageUnit("B"),
            asn=self.api_response_three["client"]["asn"],
            ip=self.api_response_three["client"]["ip"],
            isp=self.api_response_three["client"]["isp"],
            city=self.api_response_three["client"]["location"]["city"],
            country=self.api_response_three["client"]["location"]["country"],
            urlcount=3,
            reason_terminated=self.fast_data_three["reason_terminated"],
            errors=[],
        )
        self.thread_result_three_list = [
            [
                NetflixFastThreadResult(
                    id="1",
                    host="afakeurl.1.notreal.net",
                    city=self.api_response_three["targets"][0]["location"]["city"],
                    country=self.api_response_three["targets"][0]["location"][
                        "country"
                    ],
                    download_size=0,
                    download_size_unit=StorageUnit("B"),
                    download_rate=0,
                    download_rate_unit=NetworkUnit("bit/s"),
                    elapsed_time=0,
                    elapsed_time_unit=TimeUnit("s"),
                    errors=[],
                ),
                LatencyMeasurementResult(
                    id="1",
                    host="afakeurl.1.notreal.net",
                    minimum_latency=0,
                    average_latency=0,
                    maximum_latency=0,
                    median_deviation=0,
                    packets_transmitted=0,
                    packets_received=0,
                    packets_lost=0,
                    packets_lost_unit=RatioUnit("%"),
                    elapsed_time=0,
                    elapsed_time_unit=TimeUnit("d"),
                    errors=[],
                ),
            ],
            [
                NetflixFastThreadResult(
                    id="1",
                    host="afakeurl.1.notreal.net",
                    city=self.api_response_three["targets"][1]["location"]["city"],
                    country=self.api_response_three["targets"][1]["location"][
                        "country"
                    ],
                    download_size=0,
                    download_size_unit=StorageUnit("B"),
                    download_rate=0,
                    download_rate_unit=NetworkUnit("bit/s"),
                    elapsed_time=0,
                    elapsed_time_unit=TimeUnit("s"),
                    errors=[],
                ),
                LatencyMeasurementResult(
                    id="1",
                    host="afakeurl.1.notreal.net",
                    minimum_latency=0,
                    average_latency=0,
                    maximum_latency=0,
                    median_deviation=0,
                    packets_transmitted=0,
                    packets_received=0,
                    packets_lost=0,
                    packets_lost_unit=RatioUnit("%"),
                    elapsed_time=0,
                    elapsed_time_unit=TimeUnit("d"),
                    errors=[],
                ),
            ],
            [
                NetflixFastThreadResult(
                    id="1",
                    host="afakeurl.1.notreal.net",
                    city=self.api_response_three["targets"][2]["location"]["city"],
                    country=self.api_response_three["targets"][2]["location"][
                        "country"
                    ],
                    download_size=0,
                    download_size_unit=StorageUnit("B"),
                    download_rate=0,
                    download_rate_unit=NetworkUnit("bit/s"),
                    elapsed_time=0,
                    elapsed_time_unit=TimeUnit("s"),
                    errors=[],
                ),
                LatencyMeasurementResult(
                    id="1",
                    host="afakeurl.1.notreal.net",
                    minimum_latency=0,
                    average_latency=0,
                    maximum_latency=0,
                    median_deviation=0,
                    packets_transmitted=0,
                    packets_received=0,
                    packets_lost=0,
                    packets_lost_unit=RatioUnit("%"),
                    elapsed_time=0,
                    elapsed_time_unit=TimeUnit("d"),
                    errors=[],
                ),
            ],
        ]

    @mock.patch(
        "netmeasure.measurements.netflix_fast.measurements.NetflixFastMeasurement._get_fast_result"
    )
    @mock.patch(
        "netmeasure.measurements.netflix_fast.measurements.NetflixFastMeasurement._get_url_result"
    )
    def test_measure(self, mock_get_url_result, mock_get_fast_result):
        mock_get_url_result.side_effect = self.thread_result_three_list
        mock_get_fast_result.return_value = self.fast_result_three
        assert self.nft.measure() == (
            [self.fast_result_three]
            + self.thread_result_three_list[0]
            + self.thread_result_three_list[1]
            + self.thread_result_three_list[2]
        )

    @mock.patch(
        "netmeasure.measurements.netflix_fast.measurements.NetflixFastMeasurement._manage_threads"
    )
    @mock.patch(
        "netmeasure.measurements.netflix_fast.measurements.NetflixFastMeasurement._get_connection"
    )
    @mock.patch("requests.Session")
    def test_fast_result(
        self, mock_get_session, mock_get_connection, mock_manage_threads
    ):
        mock_session = mock.MagicMock()
        mock_resp = mock.MagicMock()
        mock_resp.text = '<script src="(This is the script)">'
        mock_script_resp = mock.MagicMock()
        mock_script_resp.text = (
            'This is the script response, containing token:"This is the token"'
        )
        mock_api_resp = mock.MagicMock()
        mock_api_resp.json.return_value = self.api_response_three
        mock_session.get.side_effect = [mock_resp, mock_script_resp, mock_api_resp]
        mock_get_session.return_value = mock_session
        mock_manage_threads.return_value = self.fast_data_three
        self.nft.thread_results = [
            {
                "index": i,
                "elapsed_time": None,
                "download_size": 0,
                "download_rate": 0,
                "url": None,
                "location": None,
            }
            for i in range(self.nft.urlcount)  # Generate thread results dict structure
        ]

        assert self.nft._get_fast_result() == self.fast_result_three

    @mock.patch(
        "netmeasure.measurements.netflix_fast.measurements.NetflixFastMeasurement._manage_threads"
    )
    @mock.patch(
        "netmeasure.measurements.netflix_fast.measurements.NetflixFastMeasurement._get_connection"
    )
    @mock.patch(
        "netmeasure.measurements.netflix_fast.measurements.NetflixFastMeasurement._query_api"
    )
    @mock.patch("requests.Session")
    def test_fast_result_client_data(
        self, mock_get_session, mock_query_api, mock_get_connection, mock_manage_threads
    ):
        assert True

    # def test_thread_adding_terminating(self):
    #     nft = NetflixFastMeasurement("1", urlcount=5, terminate_on_thread_complete=True)
    #     conns = []
    #     for i in range(1,10):
    #         m = mock.MagicMock()
    #         m.iter_content.return_value = ["BYTES" for iter in range(i)]
    #         conns.append(m)
    #     x = nft._manage_threads(conns)
    #     assert ((x["reason_terminated"] == "thread_complete"))

    def test_thread_adding_all_complete(self):
        nft = NetflixFastMeasurement(
            "1", urlcount=5, terminate_on_thread_complete=False
        )
        conns = []
        for i in range(1, 6):
            m = mock.MagicMock()
            m.iter_content.return_value = ["BYTES" for iter in range(i)]
            conns.append(m)
        nft.thread_results = [
            {
                "index": i,
                "elapsed_time": None,
                "download_size": 0,
                "download_rate": 0,
                "url": None,
                "location": None,
            }
            for i in range(nft.urlcount)  # Generate thread results dict structure
        ]
        x = nft._manage_threads(conns)
        assert (x["reason_terminated"] == "all_complete") & (x["total"] == 75)

    @mock.patch(
        "netmeasure.measurements.netflix_fast.measurements.NetflixFastMeasurement._is_stabilised"
    )
    def test_thread_adding_stabilised(self, mock_is_stabilised):
        mock_is_stabilised.return_value = True
        nft = NetflixFastMeasurement(
            "1",
            urlcount=5,
            terminate_on_thread_complete=False,
            terminate_on_result_stable=True,
        )
        conns = []
        for i in range(1, 6):
            m = mock.MagicMock()
            m.iter_content.return_value = ["BYTES" for _ in range(i)]
            conns.append(m)
        x = nft._manage_threads(conns)
        assert x["reason_terminated"] == "result_stabilised"

    @mock.patch(
        "netmeasure.measurements.latency.measurements.LatencyMeasurement.measure"
    )
    def test_latency_result(self, mock_latency_measure):
        mock_latency_result = LatencyMeasurementResult(
            id=("1"),
            host="afakeurl.1.notreal.net",
            minimum_latency=0,
            average_latency=0,
            maximum_latency=0,
            median_deviation=0,
            packets_transmitted=0,
            packets_received=0,
            packets_lost=0,
            packets_lost_unit=RatioUnit("%"),
            elapsed_time=0,
            elapsed_time_unit=TimeUnit("d"),
            errors=[],
        )
        mock_latency_measure.return_value = [mock_latency_result]
        mock_thread_result = {
            "url": "https://afakeurl.1.notreal.net/speedtest",
            "location": {"city": "Foreign City One", "country": "Foreign Country One"},
            "download_size": 0,
            "download_rate": 0,
            "elapsed_time": 0,
        }
        assert (
            self.nft._get_url_result(mock_thread_result)
            == self.thread_result_three_list[0]
        )


class HelperFunctionTestCase(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.nft = NetflixFastMeasurement("1")

    @mock.patch("requests.Session")
    def test_get_connection(self, mock_get_session):
        mock_session = mock.MagicMock()
        mock_session.get.side_effect = ["first_url", "second_url", "third_url"]
        mock_get_session.return_value = mock_session
        nft = NetflixFastMeasurement("1")
        assert [
            nft._get_connection("first"),
            nft._get_connection("second"),
            nft._get_connection("third"),
        ] == ["first_url", "second_url", "third_url"]
        assert len(nft.sessions) == 3

    def test_is_stabilised_is_stable(self):
        mock_elapsed_time = MIN_TIME_SECONDS + 1
        mock_percent_deltas = [
            STABLE_MEASUREMENTS_DELTA - sys.float_info.epsilon
        ] * MEASUREMENTS_COUNTED_BEFORE_CONSIDERED_STABLE
        assert self.nft._is_stabilised(mock_percent_deltas, mock_elapsed_time) is True

    def test_is_stabilised_not_stable(self):
        mock_elapsed_time = MIN_TIME_SECONDS + 1
        mock_percent_deltas = [STABLE_MEASUREMENTS_DELTA - sys.float_info.epsilon] * (
            MEASUREMENTS_COUNTED_BEFORE_CONSIDERED_STABLE - 1
        )
        mock_percent_deltas.append(STABLE_MEASUREMENTS_DELTA + sys.float_info.epsilon)
        assert self.nft._is_stabilised(mock_percent_deltas, mock_elapsed_time) is False

    def test_is_stabilised_too_short(self):
        mock_elapsed_time = MIN_TIME_SECONDS + 1
        mock_percent_deltas = [STABLE_MEASUREMENTS_DELTA - sys.float_info.epsilon] * (
            MEASUREMENTS_COUNTED_BEFORE_CONSIDERED_STABLE - 1
        )
        assert self.nft._is_stabilised(mock_percent_deltas, mock_elapsed_time) is False

    def test_is_stabilised_too_early(self):
        mock_elapsed_time = MIN_TIME_SECONDS - 1
        mock_percent_deltas = [
            STABLE_MEASUREMENTS_DELTA - sys.float_info.epsilon
        ] * MEASUREMENTS_COUNTED_BEFORE_CONSIDERED_STABLE
        assert self.nft._is_stabilised(mock_percent_deltas, mock_elapsed_time) is False


class ErrorsTestCase(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.nft = NetflixFastMeasurement("1")

    @mock.patch("requests.Session")
    def test_netflix_response_err(self, mock_get_session):
        mock_error_result = NetflixFastMeasurementResult(
            id=("1"),
            download_rate=None,
            download_rate_unit=None,
            download_size=None,
            download_size_unit=None,
            asn=None,
            ip=None,
            isp=None,
            city=None,
            country=None,
            urlcount=self.nft.urlcount,
            reason_terminated=None,
            errors=[
                Error(
                    key="netflix-response",
                    description=NETFLIX_ERRORS.get("netflix-response", ""),
                    traceback="Failed to pretend to establish a new connection",
                )
            ],
        )
        mock_session = mock.MagicMock()
        mock_session.get.side_effect = [
            ConnectionError("Failed to pretend to establish a new connection")
        ]
        mock_get_session.return_value = mock_session
        assert self.nft._get_fast_result() == mock_error_result

    @mock.patch("requests.Session")
    def test_netflix_regex_err(self, mock_get_session):
        mock_error_result = NetflixFastMeasurementResult(
            id=("1"),
            download_rate=None,
            download_rate_unit=None,
            download_size=None,
            download_size_unit=None,
            asn=None,
            ip=None,
            isp=None,
            city=None,
            country=None,
            urlcount=self.nft.urlcount,
            reason_terminated=None,
            errors=[
                Error(
                    key="netflix-script-regex",
                    description=NETFLIX_ERRORS.get("netflix-script-regex", ""),
                    traceback='<Bript Brc="(This should be impossible to find)">',
                )
            ],
        )
        mock_session = mock.MagicMock()
        mock_resp = mock.MagicMock()
        mock_resp.text = '<Bript Brc="(This should be impossible to find)">'
        mock_session.get.side_effect = [mock_resp]
        mock_get_session.return_value = mock_session
        assert self.nft._get_fast_result() == mock_error_result

    @mock.patch("requests.Session")
    def test_netflix_script_response_err(self, mock_get_session):
        mock_error_result = NetflixFastMeasurementResult(
            id=("1"),
            download_rate=None,
            download_rate_unit=None,
            download_size=None,
            download_size_unit=None,
            asn=None,
            ip=None,
            isp=None,
            city=None,
            country=None,
            urlcount=self.nft.urlcount,
            reason_terminated=None,
            errors=[
                Error(
                    key="netflix-script-response",
                    description=NETFLIX_ERRORS.get("netflix-script-response", ""),
                    traceback="Failed to pretend to establish a new connection",
                )
            ],
        )
        mock_session = mock.MagicMock()
        mock_resp = mock.MagicMock()
        mock_resp.text = '<script src="(This is the script)">'
        mock_session.get.side_effect = [
            mock_resp,
            ConnectionError("Failed to pretend to establish a new connection"),
        ]
        mock_get_session.return_value = mock_session
        assert self.nft._get_fast_result() == mock_error_result

    @mock.patch("requests.Session")
    def test_netflix_token_regex_err(self, mock_get_session):
        mock_error_result = NetflixFastMeasurementResult(
            id=("1"),
            download_rate=None,
            download_rate_unit=None,
            download_size=None,
            download_size_unit=None,
            asn=None,
            ip=None,
            isp=None,
            city=None,
            country=None,
            urlcount=self.nft.urlcount,
            reason_terminated=None,
            errors=[
                Error(
                    key="netflix-token-regex",
                    description=NETFLIX_ERRORS.get("netflix-token-regex", ""),
                    traceback='This is an invalid script response, containing BRoken:"This is the token"',
                )
            ],
        )
        mock_session = mock.MagicMock()
        mock_resp = mock.MagicMock()
        mock_resp.text = '<script src="(This is the script)">'
        mock_script_resp = mock.MagicMock()
        mock_script_resp.text = (
            'This is an invalid script response, containing BRoken:"This is the token"'
        )
        mock_session.get.side_effect = [mock_resp, mock_script_resp]
        mock_get_session.return_value = mock_session
        assert self.nft._get_fast_result() == mock_error_result

    @mock.patch("requests.Session")
    def test_netflix_api_response_err(self, mock_get_session):
        mock_error_result = NetflixFastMeasurementResult(
            id=("1"),
            download_rate=None,
            download_rate_unit=None,
            download_size=None,
            download_size_unit=None,
            asn=None,
            ip=None,
            isp=None,
            city=None,
            country=None,
            urlcount=self.nft.urlcount,
            reason_terminated=None,
            errors=[
                Error(
                    key="netflix-api-response",
                    description=NETFLIX_ERRORS.get("netflix-api-response", ""),
                    traceback="Failed to pretend to establish a new connection",
                )
            ],
        )
        mock_session = mock.MagicMock()
        mock_resp = mock.MagicMock()
        mock_resp.text = '<script src="(This is the script)">'
        mock_script_resp = mock.MagicMock()
        mock_script_resp.text = (
            'This is the script response, containing token:"This is the token"'
        )
        mock_session.get.side_effect = [
            mock_resp,
            mock_script_resp,
            ConnectionError("Failed to pretend to establish a new connection"),
        ]
        mock_get_session.return_value = mock_session
        assert self.nft._get_fast_result() == mock_error_result

    @mock.patch("requests.Session")
    def test_netflix_api_json_err(self, mock_get_session):
        mock_error_result = NetflixFastMeasurementResult(
            id=("1"),
            download_rate=None,
            download_rate_unit=None,
            download_size=None,
            download_size_unit=None,
            asn=None,
            ip=None,
            isp=None,
            city=None,
            country=None,
            urlcount=self.nft.urlcount,
            reason_terminated=None,
            errors=[
                Error(
                    key="netflix-api-json",
                    description=NETFLIX_ERRORS.get("netflix-api-json", ""),
                    traceback="Json Decode screwed up :(: line 1 column 1 (char 0)",
                )
            ],
        )
        mock_session = mock.MagicMock()
        mock_resp = mock.MagicMock()
        mock_resp.text = '<script src="(This is the script)">'
        mock_script_resp = mock.MagicMock()
        mock_script_resp.text = (
            'This is the script response, containing token:"This is the token"'
        )
        mock_api_resp = mock.MagicMock()
        mock_api_resp.json.side_effect = json.decoder.JSONDecodeError(
            "Json Decode screwed up :(", "{BuSTED_JSON ::: asdf}", 0
        )
        mock_session.get.side_effect = [mock_resp, mock_script_resp, mock_api_resp]
        mock_get_session.return_value = mock_session
        assert self.nft._get_fast_result() == mock_error_result

    @mock.patch("requests.Session")
    def test_netflix_api_parse_err(self, mock_get_session):
        mock_error_result = NetflixFastMeasurementResult(
            id=("1"),
            download_rate=None,
            download_rate_unit=None,
            download_size=None,
            download_size_unit=None,
            asn=None,
            ip=None,
            isp=None,
            city=None,
            country=None,
            urlcount=self.nft.urlcount,
            reason_terminated=None,
            errors=[
                Error(
                    key="netflix-api-parse",
                    description=NETFLIX_ERRORS.get("netflix-api-parse", ""),
                    traceback="string indices must be integers, not 'str'",
                )
            ],
        )
        # JSONDecodeError changed in 3.11
        if sys.version_info < (3, 11):
            mock_error_result = NetflixFastMeasurementResult(
                id=("1"),
                download_rate=None,
                download_rate_unit=None,
                download_size=None,
                download_size_unit=None,
                asn=None,
                ip=None,
                isp=None,
                city=None,
                country=None,
                urlcount=self.nft.urlcount,
                reason_terminated=None,
                errors=[
                    Error(
                        key="netflix-api-parse",
                        description=NETFLIX_ERRORS.get("netflix-api-parse", ""),
                        traceback="string indices must be integers",
                    )
                ],
            )
        mock_session = mock.MagicMock()
        mock_resp = mock.MagicMock()
        mock_resp.text = '<script src="(This is the script)">'
        mock_script_resp = mock.MagicMock()
        mock_script_resp.text = (
            'This is the script response, containing token:"This is the token"'
        )
        mock_api_resp = mock.MagicMock()
        mock_api_resp.json.side_effect = {"correct_values": "are not in here >:)"}
        mock_session.get.side_effect = [mock_resp, mock_script_resp, mock_api_resp]
        mock_get_session.return_value = mock_session
        assert self.nft._get_fast_result() == mock_error_result

    @mock.patch("requests.Session")
    def test_netflix_api_parse_targets_err(self, mock_get_session):
        mock_error_result = NetflixFastMeasurementResult(
            id=("1"),
            download_rate=None,
            download_rate_unit=None,
            download_size=None,
            download_size_unit=None,
            asn=None,
            ip=None,
            isp=None,
            city=None,
            country=None,
            urlcount=self.nft.urlcount,
            reason_terminated=None,
            errors=[
                Error(
                    key="netflix-api-parse",
                    description=NETFLIX_ERRORS.get("netflix-api-parse", ""),
                    traceback="0",
                )
            ],
        )
        mock_session = mock.MagicMock()
        mock_resp = mock.MagicMock()
        mock_resp.text = '<script src="(This is the script)">'
        mock_script_resp = mock.MagicMock()
        mock_script_resp.text = (
            'This is the script response, containing token:"This is the token"'
        )
        mock_api_resp = mock.MagicMock()
        mock_api_resp.json = mock.MagicMock()
        mock_api_resp.json.side_effect = [
            {"targets": {"correct_values": "are not in here >:)"}}
        ]
        mock_session.get.side_effect = [mock_resp, mock_script_resp, mock_api_resp]
        mock_get_session.return_value = mock_session
        assert self.nft._get_fast_result() == mock_error_result

    @mock.patch("requests.Session")
    def test_netflix_connection_err(self, mock_get_session):
        mock_error_result = NetflixFastMeasurementResult(
            id=("1"),
            download_rate=None,
            download_rate_unit=None,
            download_size=None,
            download_size_unit=None,
            asn=None,
            ip=None,
            isp=None,
            city=None,
            country=None,
            urlcount=self.nft.urlcount,
            reason_terminated=None,
            errors=[
                Error(
                    key="netflix-connection",
                    description=NETFLIX_ERRORS.get("netflix-connection", ""),
                    traceback="Failed to pretend to establish a new connection",
                )
            ],
        )
        mock_session = mock.MagicMock()
        mock_resp = mock.MagicMock()
        mock_resp.text = '<script src="(This is the script)">'
        mock_script_resp = mock.MagicMock()
        mock_script_resp.text = (
            'This is the script response, containing token:"This is the token"'
        )
        mock_api_resp = mock.MagicMock()
        mock_api_resp.json.side_effect = [
            {
                "client": {},
                "targets": [
                    {
                        "url": "https://afakeurl.1.notreal.net/speedtest",
                        "location": {
                            "city": "Foreign City One",
                            "country": "Foreign Country One",
                        },
                    }
                ],
            }
        ]
        mock_session.get.side_effect = [
            mock_resp,
            mock_script_resp,
            mock_api_resp,
            ConnectionError("Failed to pretend to establish a new connection"),
        ]
        mock_get_session.return_value = mock_session
        self.nft.thread_results = [
            {
                "index": i,
                "elapsed_time": None,
                "download_size": 0,
                "download_rate": 0,
                "url": None,
                "location": None,
            }
            for i in range(self.nft.urlcount)  # Generate thread results dict structure
        ]
        assert self.nft._get_fast_result() == mock_error_result



import typing
from dataclasses import dataclass

from netmeasure.measurements.base.results import MeasurementResult
from netmeasure.units import TimeUnit, StorageUnit, RatioUnit, NetworkUnit


@dataclass(frozen=True)
class IPRouteMeasurementResult(MeasurementResult):
    """Encapsulates the result from an IPRoute measurement."""

    host: typing.Optional[str]
    hop_count: typing.Optional[int]
    ip: typing.Optional[str]
    route: typing.Optional[list]



import typing
from dataclasses import dataclass

from netmeasure.measurements.base.results import MeasurementResult
from netmeasure.units import TimeUnit, StorageUnit, RatioUnit, NetworkUnit


@dataclass(frozen=True)
class YoutubeDownloadMeasurementResult(MeasurementResult):
    """Encapsulates the results from a YouTube measurement."""

    download_rate: typing.Optional[float]
    download_rate_unit: typing.Optional[NetworkUnit]
    download_size: typing.Optional[float]
    download_size_unit: typing.Optional[StorageUnit]
    url: typing.Optional[str]
    elapsed_time: typing.Optional[float]
    elapsed_time_unit: typing.Optional[TimeUnit]



# -*- coding: utf-8 -*-
from unittest import TestCase, mock
from unittest.mock import call

import six
import subprocess

from netmeasure.measurements.latency.measurements import LatencyMeasurement
from netmeasure.measurements.base.results import Error
from netmeasure.measurements.webpage_download.measurements import (
    WebpageDownloadMeasurement,
)
from netmeasure.measurements.webpage_download.measurements import WEB_ERRORS

from netmeasure.measurements.webpage_download.results import (
    WebpageDownloadMeasurementResult,
)
from netmeasure.measurements.latency.results import LatencyMeasurementResult

from netmeasure.units import NetworkUnit, StorageUnit, TimeUnit, RatioUnit


class WebpageDownloadMeasurementResultsTestCase(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.wpm = WebpageDownloadMeasurement("test", "http://validfakehost.com/test")
        self.simple_webpage_output = WebpageDownloadMeasurementResult(
            id="test",
            url="http://validfakehost.com/test",
            download_rate=100 / 1.00 * 8,
            download_rate_unit=NetworkUnit("bit/s"),
            download_size=100,
            download_size_unit=StorageUnit("B"),
            asset_count=123,
            failed_asset_downloads=0,
            elapsed_time=1.00,
            elapsed_time_unit=TimeUnit("s"),
            errors=[],
        )
        self.simple_asset_download_metrics = {
            "asset_download_size": 90,
            "failed_asset_downloads": 0,
            "completion_time": 2.00,
        }
        self.get_error_result = WebpageDownloadMeasurementResult(
            id="test",
            url="http://validfakehost.com/test",
            download_rate_unit=None,
            download_rate=None,
            download_size=None,
            download_size_unit=None,
            asset_count=None,
            failed_asset_downloads=None,
            elapsed_time=None,
            elapsed_time_unit=None,
            errors=[
                Error(
                    key="web-get",
                    description=WEB_ERRORS.get("web-get", ""),
                    traceback="[Errno -2] Name or service not known",
                )
            ],
        )

    @mock.patch(
        "netmeasure.measurements.webpage_download.measurements.WebpageDownloadMeasurement._get_webpage_result"
    )
    def test_measure(self, mock_get_webpage):
        mock_get_webpage.return_value = self.simple_webpage_output
        self.assertEqual(self.wpm.measure(), self.simple_webpage_output)

    @mock.patch(
        "netmeasure.measurements.webpage_download.measurements.WebpageDownloadMeasurement._download_assets"
    )
    @mock.patch(
        "netmeasure.measurements.webpage_download.measurements.WebpageDownloadMeasurement._parse_html"
    )
    @mock.patch("requests.Session")
    @mock.patch("time.time")
    def test_get_requests_measurement(
        self, mock_time, mock_get_session, mock_parse_html, mock_download_assets
    ):
        mock_time.return_value = 1.00
        mock_session = mock.MagicMock()
        mock_resp = mock.MagicMock()
        # Total 'downloaded' size is 100 bytes
        mock_resp.text = "Ten chars_"
        mock_session.get.side_effect = [mock_resp]
        mock_parse_html.return_value = [
            "URL number {x}".format(x=i + 1) for i in range(123)
        ]
        mock_download_assets.return_value = self.simple_asset_download_metrics
        mock_get_session.return_value = mock_session
        self.assertEqual(
            self.wpm._get_webpage_result(
                "http://validfakehost.com/test", "validfakehost.com", "https"
            ),
            self.simple_webpage_output,
        )

    @mock.patch(
        "netmeasure.measurements.webpage_download.measurements.WebpageDownloadMeasurement._download_assets"
    )
    @mock.patch(
        "netmeasure.measurements.webpage_download.measurements.WebpageDownloadMeasurement._parse_html"
    )
    @mock.patch("requests.Session")
    @mock.patch("time.time")
    def test_get_requests_error(
        self, mock_time, mock_get_session, mock_parse_html, mock_download_assets
    ):
        mock_time.return_value = 1.00
        mock_session = mock.MagicMock()
        mock_session.get.side_effect = [
            ConnectionError("[Errno -2] Name or service not known")
        ]
        mock_get_session.return_value = mock_session
        self.assertEqual(
            self.wpm._get_webpage_result(
                "http://validfakehost.com/test", "validfakehost.com", "https"
            ),
            self.get_error_result,
        )


class WebpageHTMLParseTestCase(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.wpm = WebpageDownloadMeasurement("test", "http://validfakehost.com/test")
        self.img_html = (
            '<img alt="HBOX" id="logo" src="/logo/src/name.png"/>\n'
            '<img alt="HBOX" id="another" src="/another/src/name.png"/>\n'
            '<img alt="HBOX" id="ayylmao" src="/ayylmao/src/name.png"/>\n'
            '<img alt="HBOX" id="no source"/>\n'
        )
        self.img_urls = [
            "/logo/src/name.png",
            "/another/src/name.png",
            "/ayylmao/src/name.png",
        ]
        self.script_html = (
            '<script src="/script_one/src/name.js"/>\n'
            '<script src="/script_two/src/name.js"/>\n'
            '<script src="/script_three/src/name.js"/>\n'
            "<script/>\n"
        )
        self.script_urls = [
            "/script_one/src/name.js",
            "/script_two/src/name.js",
            "/script_three/src/name.js",
        ]
        self.link_html = (
            '<link href="http://validfakehost.com/test" rel="manifest"/>\n'
            '<link href="http://externalfakehost.com/one" rel="modulepreload"/>\n'
            '<link href="http://externalfakehost.com/two" rel="preload"/>\n'
            '<link href="http://externalfakehost.com/three" rel="prerender"/>\n'
            '<link href="http://externalfakehost.com/four.css" rel="stylesheet"/>\n'
            '<link href="http://externalfakehost.com/five.png" rel="apple-touch-icon"/>\n'
            '<link href="http://externalfakehost.com/six.ico" rel="icon"/>\n'
            '<link href="http://externalfakehost.com/seven.ico" rel="shortcut icon"/>\n'
            '<link href="http://externalfakehost.com/eight.txt" rel="dont care"/>\n'
            '<link href="http://externalfakehost.com/no_rel.txt"/>\n'
            '<link rel="icon"/>\n'
        )
        self.link_urls = [
            "http://validfakehost.com/test",
            "http://externalfakehost.com/one",
            "http://externalfakehost.com/two",
            "http://externalfakehost.com/three",
            "http://externalfakehost.com/four.css",
            "http://externalfakehost.com/five.png",
            "http://externalfakehost.com/six.ico",
            "http://externalfakehost.com/seven.ico",
        ]

    def test_parse_img(self):
        self.assertEqual(self.wpm._parse_html(self.img_html), self.img_urls)

    def test_parse_script(self):
        self.assertEqual(self.wpm._parse_html(self.script_html), self.script_urls)

    def test_parse_links(self):
        self.assertEqual(self.wpm._parse_html(self.link_html), self.link_urls)


class WebpageAssetDownloadTestCase(TestCase):
    def setUp(self) -> None:
        super().setUp()
        self.wpm = WebpageDownloadMeasurement("test", "http://validfakehost.com/test")
        self.all_url_types = [
            "https://validfakehost.com/an_image.jpg",
            "/resources/client/a_stylesheet.css",
            "//res.validfakehost.com/fonts/a_font.woff2",
        ]
        self.all_success_urls_transformed = [
            call(
                "https://validfakehost.com/an_image.jpg",
                timeout=self.wpm.download_timeout,
            ),
            call(
                "https://validfakehost.com/resources/client/a_stylesheet.css",
                timeout=self.wpm.download_timeout,
            ),
            call(
                "https://res.validfakehost.com/fonts/a_font.woff2",
                timeout=self.wpm.download_timeout,
            ),
        ]
        self.all_success_dict = {
            "asset_download_size": 3,
            "failed_asset_downloads": 0,
            "completion_time": 1.23,
        }
        self.one_failure_dict = {
            "asset_download_size": 2,
            "failed_asset_downloads": 1,
            "completion_time": 1.23,
        }
        self.all_failure_dict = {
            "asset_download_size": 0,
            "failed_asset_downloads": 3,
            "completion_time": 1.23,
        }

    @mock.patch("time.time")
    def test_all_success_dict(self, mock_time):
        mock_time.return_value = 1.23
        mock_session = mock.MagicMock()
        responses = []
        for i in range(3):
            response = mock.MagicMock()
            response.text = str("i")
            response.status_code = 200
            responses.append(response)
        mock_session.get.side_effect = responses
        self.assertEqual(
            self.wpm._download_assets(
                mock_session, self.all_url_types, "validfakehost.com", "https"
            ),
            self.all_success_dict,
        )

    def test_all_success_urls(self):
        mock_session = mock.MagicMock()
        responses = []
        for i in range(3):
            response = mock.MagicMock()
            response.text = str("i")
            response.status_code = 200
            responses.append(response)
        mock_session.get.side_effect = responses
        self.wpm._download_assets(
            mock_session, self.all_url_types, "validfakehost.com", "https"
        )
        mock_session.get.assert_has_calls(self.all_success_urls_transformed)

    @mock.patch("time.time")
    def test_single_failure_code(self, mock_time):
        mock_time.return_value = 1.23
        mock_session = mock.MagicMock()
        responses = []
        for i in range(2):
            response = mock.MagicMock()
            response.text = str("i")
            response.status_code = 200
            responses.append(response)
        fail_response = mock.MagicMock()
        fail_response.text = "0"
        fail_response.status_code = 404
        responses.append(fail_response)
        mock_session.get.side_effect = responses
        self.assertEqual(
            self.wpm._download_assets(
                mock_session, self.all_url_types, "validfakehost.com", "https"
            ),
            self.one_failure_dict,
        )

    @mock.patch("time.time")
    def test_single_failure_error(self, mock_time):
        mock_time.return_value = 1.23
        mock_session = mock.MagicMock()
        responses = []
        for i in range(2):
            response = mock.MagicMock()
            response.text = str("i")
            response.status_code = 200
            responses.append(response)
        responses.append(ConnectionError)
        mock_session.get.side_effect = responses
        self.assertEqual(
            self.wpm._download_assets(
                mock_session, self.all_url_types, "validfakehost.com", "https"
            ),
            self.one_failure_dict,
        )

    @mock.patch("time.time")
    def test_all_failure(self, mock_time):
        mock_time.return_value = 1.23
        mock_session = mock.MagicMock()
        responses = []
        for i in range(2):
            response = mock.MagicMock()
            response.text = str("i")
            response.status_code = 404
            responses.append(response)
        responses.append(ConnectionError)
        mock_session.get.side_effect = responses
        self.assertEqual(
            self.wpm._download_assets(
                mock_session, self.all_url_types, "validfakehost.com", "https"
            ),
            self.all_failure_dict,
        )
